
# import json
# import os

# import pandas as pd
# import pytest

# import rcv_cruncher.contests as contests
# import rcv_cruncher.misc_tabulation as misc_tabulation
# import rcv_cruncher.rcv_base as rcv_base
# import rcv_cruncher.write_out as write_out


# def read_test_config(test_config_path):

#     # if not present, skip
#     if not os.path.isfile(test_config_path):
#         return {}

#     with open(test_config_path) as test_config_file:
#         test_config = json.load(test_config_file)

#     return test_config


# def test_tabulation(test_root_path, ballot_path):

#     # check for skips
#     # load test config
#     test_config_path = os.path.normpath(f'{ballot_path}/input/test_config.json')

#     # check for config option
#     if not read_test_config(test_config_path).get('test_tabulation', False):
#         pytest.skip(f'"test_tabulation" not set to "true" in test_config.json at {test_config_path}')

#     # proceed with test
#     contest_set_path = os.path.normpath(f'{ballot_path}/input')
#     computed_output_path = os.path.normpath(f'{ballot_path}/computed')

#     if not os.path.isdir(computed_output_path):
#         os.mkdir(computed_output_path)

#     # generate computed results
#     contest_set, _ = contests.read_contest_set(contest_set_path, override_cvr_root_dir=test_root_path)
#     rcv_obj = rcv_base.RCV.run_rcv(contest_set[0])
#     computed_rbrs = misc_tabulation.round_by_round(rcv_obj)

#     write_out.write_converted_cvr_annotated(rcv_obj, computed_output_path)

#     # compare to expected
#     for iTab, computed_rbr in enumerate(computed_rbrs, start=1):

#         # filter count columns
#         keep_cols = ['candidate'] + [col for col in computed_rbr.columns if 'count' in col]
#         computed_rbr_simple = computed_rbr.loc[:, keep_cols]

#         # write out computed for easier comparison when test fails
#         computed_rbr.to_csv(f'{computed_output_path}/round_by_round_tab{iTab}_complete.csv', index=False)
#         computed_rbr_simple.to_csv(f'{computed_output_path}/round_by_round_tab{iTab}_simplified.csv', index=False)

#         # also write out ballot allocation to help debug tests

#         # read expected tabulation
#         expected_tabulation_path = os.path.normpath(f'{ballot_path}/expected/tabulation{iTab}.csv')
#         if not os.path.isfile(expected_tabulation_path):
#             raise FileNotFoundError('testing error - contest requires multiple tabulations,'
#                                     f' but expected tabulation file {expected_tabulation_path} is not found.')
#         expected_rbr = pd.read_csv(expected_tabulation_path)

#         # tables are weird to compare
#         # convert tables to sets to check for equality
#         computed_compare = set(tuple(i) for i in computed_rbr_simple.values.tolist())
#         expected_compare = set(tuple(i) for i in expected_rbr.values.tolist())

#         assert computed_compare == expected_compare



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mrcv_cruncher` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``rcv_cruncher.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``rcv_cruncher.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import argparse
import os

import rcv_cruncher.batch as batch


def main():

    # argument parse and valid
    p = argparse.ArgumentParser(description='Analyze RCV election data.'
                                'For information on input file preparation, check the documentation at ***docs link here ****')

    p.add_argument('contest_set_path', help="Path to directory containing contest_set.csv and run_config.json.")
    p.add_argument('--fresh', action='store_true',
                   help='Delete existing results/ and converted_cvr/ directories located in contest set directory')
    # p.add_argument('--output_path', help='By default all output will be written to contest_set_path,'
    #                                      'provide this argument to specify an alternative.')

    args = p.parse_args()
    contest_set_path = args.contest_set_path
    fresh = args.fresh
    output_path = contest_set_path  # args.output_path if args.output_path else args.contest_set_path

    if not os.path.isabs(contest_set_path):
        contest_set_path = f'{os.getcwd()}/{contest_set_path}'

    if not os.path.isdir(contest_set_path):
        raise RuntimeError(f'invalid path [contest_set_path]: {contest_set_path}')

    # if not os.path.isabs(output_path):
    #     output_path = f'{os.getcwd()}/{output_path}'

    # if not os.path.isdir(output_path):
    #     raise RuntimeError(f'invalid path [output_path]: {output_path}')

    # read in contest set info
    contest_set, run_config = batch.read_contest_set(contest_set_path)

    # analyze contests
    batch.crunch_contest_set(contest_set, run_config, output_path, fresh_output=fresh)

    return(0)




import pytest

from rcv_cruncher.marks import BallotMarks

params = [
    (TypeError, 1),
    (TypeError, 1.5),
    (TypeError, 'a'),
    (TypeError, True)
]


@pytest.mark.parametrize("error_type, input", params)
def test_constructor_errors(error_type, input):

    with pytest.raises(error_type):
        BallotMarks(input)


@pytest.mark.parametrize("error_type, input", params)
def test_remove_duplicate_marks_errors(error_type, input):

    with pytest.raises(error_type):
        BallotMarks.remove_duplicate_candidate_marks(input)


@pytest.mark.parametrize("error_type, input", params)
def test_combine_writein_marks_errors(error_type, input):

    with pytest.raises(error_type):
        BallotMarks.combine_writein_marks(input)


params = [
    (TypeError, 1, 1),
    (TypeError, 1, 1.5),
    (TypeError, 1, 'a'),
    (TypeError, 1, False),
    (TypeError, 1.5, 1),
    (TypeError, 1.5, 1.5),
    (TypeError, 1.5, 'a'),
    (TypeError, 1.5, True),
    (TypeError, 'a', 1),
    (TypeError, 'a', 1.5),
    (TypeError, 'a', 'a'),
    (TypeError, 'a', True),
    (TypeError, True, 1),
    (TypeError, True, 1.5),
    (TypeError, True, 'a'),
    (TypeError, True, True)
]


@pytest.mark.parametrize("error_type, input1, input2", params)
def test_remove_mark_errors(error_type, input1, input2):

    with pytest.raises(error_type):
        BallotMarks.remove_mark(input1, input2)


def test_apply_rules_errors():

    with pytest.raises(RuntimeError):
        b = BallotMarks(['A', 'B', 'C'])
        b.apply_rules()
        b.apply_rules()




import pandas as pd

import rcv_cruncher.util as util

from rcv_cruncher.marks import BallotMarks


class CastVoteRecord_tables:

    # EXPORT

    def cvr_table(self, table_format: str = "rank") -> pd.DataFrame:

        if table_format != "rank" and table_format != "candidate":
            raise RuntimeError('table_format argument must be "rank" or "candidate"')

        if table_format == "rank":
            return self._rank_header_cvr()
        elif table_format == "candidate":
            return self._candidate_header_cvr()

    def _rank_header_cvr(self) -> pd.DataFrame:

        ballot_dict = {k: v for k, v in self.get_cvr_dict().items()}
        bs = [b.marks for b in ballot_dict['ballot_marks']]
        weight = ballot_dict['weight']
        del ballot_dict['ballot_marks']
        del ballot_dict['weight']

        # how many ranks?
        num_ranks = max(len(i) for i in bs)

        # shift over all ranks
        # bs = [[m for m in b if m != BallotMarks.SKIPPED] for b in bs]

        # make sure all ballots are lists of equal length, adding trailing 'skipped' if necessary
        bs = [b + ([BallotMarks.SKIPPED] * (num_ranks - len(b))) for b in bs]

        # assemble output_table, start with extras
        output_df = pd.DataFrame.from_dict(ballot_dict)

        # are weights all one, then dont add to output
        if not all([i == 1 for i in weight]):
            output_df['weight'] = [float(w) for w in weight]

        # add in rank columns
        for i in range(1, num_ranks + 1):
            output_df['rank' + str(i)] = [b[i-1] for b in bs]

        # add in rank columns
        # for i in range(1, num_ranks + 1):
        #    output_df[str(i)] = [''.join(b[i-1].split(' ')) for b in bs]
        #    output_df[str(i)] = output_df[str(i)].replace(to_replace=BallotMarks.SKIPPED, value='')

        return output_df

    def _candidate_header_cvr(self) -> pd.DataFrame:

        # get ballots and candidates
        ballot_dl = {k: v for k, v in self.get_cvr_dict().items()}
        candidates = self.get_candidates().unique_candidates.copy()
        candidates.update({BallotMarks.OVERVOTE})

        # remove weights if all equal to 1
        if set(ballot_dl['weight']) == {1}:
            del ballot_dl['weight']

        # add rank limit
        ballot_dl['rank_limit'] = [len(ballot_dl['ballot_marks'][0].marks)] * len(ballot_dl['ballot_marks'])

        # convert dict of list to list of dicts
        ballot_ld = util.DL2LD(ballot_dl)

        # add candidate index information
        for b in ballot_ld:
            b.update({f"candidate_{cand}": None for cand in candidates})
            for rank_idx, cand in enumerate(b['ballot_marks'].marks, start=1):
                if cand != BallotMarks.SKIPPED:
                    if not b[f"candidate_{cand}"]:
                        b[f"candidate_{cand}"] = str(rank_idx)
                    else:
                        b[f"candidate_{cand}"] += f',{rank_idx}'
            del b['ballot_marks']

        df = pd.DataFrame(ballot_ld)
        return df.reindex(sorted(df.columns), axis=1)

    def rank_usage_table(self):
        """
        DOES NOT USE BALLOT WEIGHTS
        """

        # get candidate set
        candidate_set = self.get_candidates()
        candidate_set = BallotMarks.combine_writein_marks(candidate_set)
        # candidate_set = BallotMarks.remove_mark(candidate_set, [BallotMarks.WRITEIN])
        candidate_set = sorted(candidate_set.unique_candidates)

        # get ballots
        ballot_set = [{'ballot_marks': bm, 'weight': weight}
                      for bm, weight in zip(self.get_cvr_dict()['ballot_marks'], self.get_cvr_dict()['weight'])]
        # remove skipped ranks
        ballot_set = [{'ballot_marks': BallotMarks.remove_mark(b['ballot_marks'], [BallotMarks.SKIPPED]),
                       'weight': b['weight']}
                      for b in ballot_set]
        # remove empty ballots and those that start with overvote
        ballot_set = [b for b in ballot_set
                      if len(b['ballot_marks'].marks) >= 1 and b['ballot_marks'].marks[0] != BallotMarks.OVERVOTE]
        # combine writeins
        ballot_set = [{'ballot_marks': BallotMarks.combine_writein_marks(b['ballot_marks']),
                       'weight': b['weight']}
                      for b in ballot_set]
        # remove other overvotes
        ballot_set = [{'ballot_marks': BallotMarks.remove_mark(b['ballot_marks'], [BallotMarks.OVERVOTE]),
                       'weight': b['weight']}
                      for b in ballot_set]
        # remove duplicate rankings
        ballot_set = [{'ballot_marks': BallotMarks.remove_duplicate_candidate_marks(b['ballot_marks']),
                       'weight': b['weight']}
                      for b in ballot_set]

        all_ballots_label = "Any candidate"
        n_ballots_label = "Number of Ballots (excluding undervotes and ballots with first round overvote)"
        mean_label = "Mean Valid Rankings Used (excluding duplicates)"
        # median_label = "Median Valid Rankings Used (excluding duplicates)"

        rows = [all_ballots_label] + candidate_set
        cols = [n_ballots_label, mean_label]  #, median_label]
        df = pd.DataFrame(index=rows, columns=cols)
        df.index.name = "Ballots with first choice:"

        ballot_total = sum(b['weight'] for b in ballot_set)
        mean_rankings = sum(len(b['ballot_marks'].marks) * b['weight'] for b in ballot_set) / ballot_total
        # median_rankings = statistics.median(len(b) for b in ballot_set)

        df.loc[all_ballots_label, n_ballots_label] = ballot_total
        df.loc[all_ballots_label, mean_label] = mean_rankings
        # df.loc[all_ballots_label, median_label] = median_rankings

        # group ballots by first choice
        first_choices = {cand: [] for cand in candidate_set}
        for b in ballot_set:
            first_choices[b['ballot_marks'].marks[0]].append(b)

        for cand in candidate_set:
            first_choice_ballot_total = sum(b['weight'] for b in first_choices[cand])
            df.loc[cand, n_ballots_label] = first_choice_ballot_total
            if first_choices[cand]:
                df.loc[cand, mean_label] = sum(len(b['ballot_marks'].marks) * b['weight']
                                               for b in first_choices[cand]) / first_choice_ballot_total
                # df.loc[cand, median_label] = statistics.median(len(b) for b in first_choices[cand])
            else:
                df.loc[cand, mean_label] = 0
                # df.loc[cand, median_label] = 0

        return df.applymap(util.decimal2float)

    def crossover_tables(self):

        # get candidate set
        candidate_set = self.get_candidates()
        candidate_set = BallotMarks.combine_writein_marks(candidate_set)
        # candidate_set = BallotMarks.remove_mark(candidate_set, [BallotMarks.WRITEIN])
        candidate_set = sorted(candidate_set.unique_candidates)

        ballot_dict = self.get_cvr_dict()
        ballot_weights = ballot_dict['weight']
        ballot_ranks = [BallotMarks.remove_mark(b, [BallotMarks.SKIPPED]) for b in ballot_dict['ballot_marks']]
        ballot_ranks = [BallotMarks.combine_writein_marks(b) for b in ballot_ranks]
        ballot_set = [{'ballot_marks': ranks.marks, 'weight': weight}
                      for ranks, weight in zip(ballot_ranks, ballot_weights)]

        index_label = "Ballots with first choice:"
        n_ballots_label = "Number of Ballots"

        colname_dict = {cand: cand + " ranked in top 3" for cand in candidate_set}

        rows = candidate_set
        cols = [n_ballots_label] + list(colname_dict.values())
        count_df = pd.DataFrame(index=rows, columns=cols)
        count_df.index.name = index_label
        percent_df = pd.DataFrame(index=rows, columns=cols)
        percent_df.index.name = index_label

        # group ballots by first choice
        first_choices = {cand: [] for cand in candidate_set}
        for b in ballot_set:
            if len(b['ballot_marks']) >= 1 and b['ballot_marks'][0] != BallotMarks.OVERVOTE:
                first_choices[b['ballot_marks'][0]].append(b)

        for cand in candidate_set:

            n_first_choice = sum(b['weight'] for b in first_choices[cand])
            count_df.loc[cand, n_ballots_label] = n_first_choice
            percent_df.loc[cand, n_ballots_label] = n_first_choice

            for opponent in candidate_set:

                if n_first_choice:
                    crossover_ballots = [True if opponent in b['ballot_marks'][0:min(3, len(b['ballot_marks']))] else False
                                         for b in first_choices[cand]]
                    crossover_val = sum(b['weight'] for b, flag in zip(first_choices[cand], crossover_ballots) if flag)
                    count_df.loc[cand, colname_dict[opponent]] = crossover_val
                    percent_df.loc[cand, colname_dict[opponent]] = crossover_val*100/n_first_choice
                else:
                    count_df.loc[cand, colname_dict[opponent]] = 0
                    percent_df.loc[cand, colname_dict[opponent]] = 0

        # convert decimal to float
        count_df = count_df.astype(float).round(3)
        percent_df = percent_df.astype(float).round(3)

        return count_df, percent_df

    def annotated_cvr_table(self):
        dfs = [self._rank_header_cvr(), self._cvr_stat_table]
        return pd.concat(dfs, axis='columns', sort=False)





import pytest

from rcv_cruncher.marks import BallotMarks


def test_update_marks():

    marks = ['A', 'B', 'B', BallotMarks.OVERVOTE, BallotMarks.SKIPPED]
    unique_marks = {'A', 'B', BallotMarks.OVERVOTE, BallotMarks.SKIPPED}
    unique_candidates = {'A', 'B'}

    b = BallotMarks()
    b.update_marks(marks)

    assert b.marks == marks
    assert b.unique_marks == unique_marks
    assert b.unique_candidates == unique_candidates


param_dicts = [
    ({
        'input': 'A',
        'expected': False
    }),
    ({
        'input': BallotMarks.WRITEIN,
        'expected': True
    }),
    ({
        'input': 'writein10',
        'expected': True
    }),
    ({
        'input': 'Fuwi',
        'expected': False
    }),
    ({
        'input': 'UwI',
        'expected': True
    })
]


@pytest.mark.parametrize("param_dict", param_dicts)
def test_check_writein_match(param_dict):

    computed = BallotMarks.check_writein_match(param_dict['input'])
    expected = param_dict['expected']
    assert expected == computed


param_dicts = [
    ({
        'input': ['A', 'B', 'WriteIn', 'writein10'],
        'expected': {
            'marks': ['A', 'B', BallotMarks.WRITEIN, BallotMarks.WRITEIN],
            'unique_marks': {'A', 'B', BallotMarks.WRITEIN},
            'unique_candidates': {'A', 'B', BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': ['A', 'B', 'Tuwi', 'UWI', 'uwi'],
        'expected': {
            'marks': ['A', 'B', 'Tuwi', BallotMarks.WRITEIN, BallotMarks.WRITEIN],
            'unique_marks': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN},
            'unique_candidates': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN}
        }
    })
]


@pytest.mark.parametrize("param_dict", param_dicts)
def test_combine_writein_marks(param_dict):

    b = BallotMarks(param_dict['input'])
    computed = BallotMarks.combine_writein_marks(b)

    assert computed.marks == param_dict['expected']['marks']
    assert computed.unique_marks == param_dict['expected']['unique_marks']
    assert computed.unique_candidates == param_dict['expected']['unique_candidates']


param_dicts = [
    ({
        'input': ['A', 'B', BallotMarks.WRITEIN, BallotMarks.WRITEIN],
        'expected': {
            'marks': ['A', 'B', BallotMarks.WRITEIN],
            'unique_marks': {'A', 'B', BallotMarks.WRITEIN},
            'unique_candidates': {'A', 'B', BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': ['A', 'B', 'B', 'C', 'C'],
        'expected': {
            'marks': ['A', 'B', 'C'],
            'unique_marks': {'A', 'B', 'C'},
            'unique_candidates': {'A', 'B', 'C'}
        }
    })
]


@pytest.mark.parametrize("param_dict", param_dicts)
def test_remove_duplicate_marks(param_dict):

    b = BallotMarks(param_dict['input'])
    computed = BallotMarks.remove_duplicate_candidate_marks(b)

    assert computed.marks == param_dict['expected']['marks']
    assert computed.unique_marks == param_dict['expected']['unique_marks']
    assert computed.unique_candidates == param_dict['expected']['unique_candidates']


param_dicts = [
    ({
        'input': {
            'marks': ['A', 'B', BallotMarks.WRITEIN, BallotMarks.WRITEIN],
            'remove_marks': ['A']
        },
        'expected': {
            'marks': ['B', BallotMarks.WRITEIN, BallotMarks.WRITEIN],
            'unique_marks': {'B', BallotMarks.WRITEIN},
            'unique_candidates': {'B', BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': {
            'marks': ['A', 'B', BallotMarks.WRITEIN, BallotMarks.WRITEIN],
            'remove_marks': ['A', 'B']
        },
        'expected': {
            'marks': [BallotMarks.WRITEIN, BallotMarks.WRITEIN],
            'unique_marks': {BallotMarks.WRITEIN},
            'unique_candidates': {BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': {
            'marks': ['A', 'B', BallotMarks.WRITEIN, BallotMarks.WRITEIN],
            'remove_marks': []
        },
        'expected': {
            'marks': ['A', 'B', BallotMarks.WRITEIN, BallotMarks.WRITEIN],
            'unique_marks': {'A', 'B', BallotMarks.WRITEIN},
            'unique_candidates': {'A', 'B', BallotMarks.WRITEIN}
        }
    }),
]


@pytest.mark.parametrize("param_dict", param_dicts)
def test_remove_mark(param_dict):

    b = BallotMarks(param_dict['input']['marks'])
    computed = BallotMarks.remove_mark(b, param_dict['input']['remove_marks'])

    assert computed.marks == param_dict['expected']['marks']
    assert computed.unique_marks == param_dict['expected']['unique_marks']
    assert computed.unique_candidates == param_dict['expected']['unique_candidates']


param_dicts = [
    ({
        'input': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set()
        },
        'expected': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'unique_marks': {'A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE},
            'unique_candidates': {'A', 'B', 'writein10', 'Tuwi', 'uwi', BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(combine_writein_marks=True)
        },
        'expected': {
            'marks': ['A', 'B', 'B', BallotMarks.WRITEIN, 'Tuwi', BallotMarks.WRITEIN, BallotMarks.WRITEIN, BallotMarks.OVERVOTE],
            'unique_marks': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN, BallotMarks.OVERVOTE},
            'unique_candidates': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exclude_duplicate_candidate_marks=True)
        },
        'expected': {
            'marks': ['A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'unique_marks': {'A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE},
            'unique_candidates': {'A', 'B', 'writein10', 'Tuwi', 'uwi', BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exclude_duplicate_candidate_marks=True, combine_writein_marks=True)
        },
        'expected': {
            'marks': ['A', 'B', BallotMarks.WRITEIN, 'Tuwi', BallotMarks.OVERVOTE],
            'unique_marks': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN, BallotMarks.OVERVOTE},
            'unique_candidates': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exclude_writein_marks=True)
        },
        'expected': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', 'uwi', BallotMarks.OVERVOTE],
            'unique_marks': {'A', 'B', 'writein10', 'Tuwi', 'uwi', BallotMarks.OVERVOTE},
            'unique_candidates': {'A', 'B', 'writein10', 'Tuwi', 'uwi'}
        }
    }),
    ({
        'input': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exclude_writein_marks=True, combine_writein_marks=True)
        },
        'expected': {
            'marks': ['A', 'B', 'B', 'Tuwi', BallotMarks.OVERVOTE],
            'unique_marks': {'A', 'B', 'Tuwi', BallotMarks.OVERVOTE},
            'unique_candidates': {'A', 'B', 'Tuwi'}
        }
    }),
    ({
        'input': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exclude_overvote_marks=True)
        },
        'expected': {
            'marks': ['A', 'B', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi'],
            'unique_marks': {'A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi'},
            'unique_candidates': {'A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi'}
        }
    }),
    ({
        'input': {
            'marks': [BallotMarks.SKIPPED, 'A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exclude_skipped_marks=True)
        },
        'expected': {
            'marks': ['A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'unique_marks': {'A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE},
            'unique_candidates': {'A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi'}
        }
    }),
    ({
        'input': {
            'marks': [BallotMarks.SKIPPED, 'A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(combine_writein_marks=True, exclude_duplicate_candidate_marks=True)
        },
        'expected': {
            'marks': [BallotMarks.SKIPPED, 'A', 'B', BallotMarks.WRITEIN, 'Tuwi', BallotMarks.OVERVOTE],
            'unique_marks': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN, BallotMarks.SKIPPED, BallotMarks.OVERVOTE},
            'unique_candidates': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': {
            'marks': [BallotMarks.SKIPPED, 'A', 'B', 'writein10', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(combine_writein_marks=True,
                                              exhaust_on_duplicate_candidate_marks=True,
                                              treat_combined_writeins_as_exhaustable_duplicates=True)
        },
        'expected': {
            'marks': [BallotMarks.SKIPPED, 'A', 'B', BallotMarks.WRITEIN, 'Tuwi'],
            'unique_marks': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN, BallotMarks.SKIPPED},
            'unique_candidates': {'A', 'B', 'Tuwi', BallotMarks.WRITEIN}
        }
    }),
    ({
        'input': {
            'marks': [BallotMarks.SKIPPED, 'A', 'B', 'B', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exhaust_on_duplicate_candidate_marks=True)
        },
        'expected': {
            'marks': [BallotMarks.SKIPPED, 'A', 'B'],
            'unique_marks': {'A', 'B', BallotMarks.SKIPPED},
            'unique_candidates': {'A', 'B'}
        }
    }),
    ({
        'input': {
            'marks': [BallotMarks.OVERVOTE, 'A', 'B', 'B', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exhaust_on_overvote_marks=True)
        },
        'expected': {
            'marks': [],
            'unique_marks': set(),
            'unique_candidates': set()
        }
    }),
    ({
        'input': {
            'marks': [BallotMarks.SKIPPED, 'A', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exhaust_on_repeated_skipped_marks=True)
        },
        'expected': {
            'marks': [BallotMarks.SKIPPED, 'A', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE],
            'unique_marks': {BallotMarks.SKIPPED, 'A', 'Tuwi', BallotMarks.WRITEIN, 'uwi', BallotMarks.OVERVOTE},
            'unique_candidates': {'A', 'Tuwi', BallotMarks.WRITEIN, 'uwi'}
        }
    }),
    ({
        'input': {
            'marks': ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'Tuwi', 'uwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exhaust_on_repeated_skipped_marks=True)
        },
        'expected': {
            'marks': ['A'],
            'unique_marks': {'A'},
            'unique_candidates': {'A'}
        }
    }),
    ({
        'input': {
            'marks': ['uwi', 'A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'Tuwi', BallotMarks.OVERVOTE],
            'rules': BallotMarks.new_rule_set(exclude_duplicate_candidate_marks=True,
                                              exclude_overvote_marks=True,
                                              exclude_skipped_marks=True,
                                              exclude_writein_marks=True,
                                              combine_writein_marks=True,
                                              treat_combined_writeins_as_exhaustable_duplicates=True,
                                              exhaust_on_duplicate_candidate_marks=True,
                                              exhaust_on_overvote_marks=True,
                                              exhaust_on_repeated_skipped_marks=True)
        },
        'expected': {
            'marks': ['A'],
            'unique_marks': {'A'},
            'unique_candidates': {'A'}
        }
    })
]


@pytest.mark.parametrize("param_dict", param_dicts)
def test_apply_rules(param_dict):

    b = BallotMarks(param_dict['input']['marks'])
    b.apply_rules(**param_dict['input']['rules'])

    assert b.marks == param_dict['expected']['marks']
    assert b.unique_marks == param_dict['expected']['unique_marks']
    assert b.unique_candidates == param_dict['expected']['unique_candidates']




from typing import (Optional, Union)

import pathlib
import json
import itertools
import pandas as pd

from rcv_cruncher.util import NAN
from rcv_cruncher.marks import BallotMarks
import rcv_cruncher.util as util


class RCV_tables:

    def condorcet_tables(self):
        """
        Returns a two condorcet tables as a pandas data frame with candidate names as row and column indices.
        One contains counts and the other contains percents.

        Each cell indicates the count/percentage of ballots that ranked the row-candidate over
        the column-candidate (including ballots that only ranked the row-candidate). When calculating percents,
        the denominator is each cell is the number of ballots that ranked the row-candidate OR the column-candidate.

        Symmetric cells about the diagonal should sum to 100 (for the percent table).
        """

        candidate_set = self.get_candidates()
        candidate_set = BallotMarks.combine_writein_marks(candidate_set)
        # candidate_set = BallotMarks.remove_mark(candidate_set, BallotMarks.WRITEIN)
        candidate_set = sorted(candidate_set.unique_candidates)

        cleaned_dict = self.get_cvr_dict(self._contest_rule_set_name)
        ballot_set = [{'ranks': ranks.marks, 'weight': weight}
                      for ranks, weight in zip(cleaned_dict['ballot_marks'], cleaned_dict['weight'])]

        # create data frame that will be populated and output
        condorcet_percent_df = pd.DataFrame(util.NAN, index=candidate_set, columns=candidate_set)
        condorcet_count_df = pd.DataFrame(util.NAN, index=candidate_set, columns=candidate_set)

        # turn ballot-lists into ballot-dict with
        # key 'id' containing a unique integer id for the ballot
        # key 'ranks' containing the original ballot-list
        ballot_dicts = [{'id': ind, 'ballot': ballot} for ind, ballot in enumerate(ballot_set)]

        # make dictionary with candidate as key, and value as list of ballot-dicts
        # that contain their name in any rank
        cand_ballot_dict = {cand: [ballot for ballot in ballot_dicts if cand in ballot['ballot']['ranks']]
                            for cand in candidate_set}

        # all candidate pairs
        cand_pairs = itertools.combinations(candidate_set, 2)

        for pair in cand_pairs:
            cand1 = pair[0]
            cand2 = pair[1]

            # get the union of their ballots
            combined_ballot_list = cand_ballot_dict[cand1] + cand_ballot_dict[cand2]
            uniq_pair_ballots = list({v['id']: v['ballot'] for v in combined_ballot_list}.values())

            uniq_pair_ballots_weights = [ballot['weight'] for ballot in uniq_pair_ballots]
            sum_weighted_ballots = sum(uniq_pair_ballots_weights)

            # which ballots rank cand1 above cand2?
            cand1_vs_cand2 = [util.index_inf(b['ranks'], cand1) < util.index_inf(b['ranks'], cand2) for b in uniq_pair_ballots]
            cand1_vs_cand2_weightsum = sum(weight * flag for weight, flag
                                           in zip(uniq_pair_ballots_weights, cand1_vs_cand2))

            # the remainder then must rank cand2 over cand1
            cand2_vs_cand1 = [not i for i in cand1_vs_cand2]
            cand2_vs_cand1_weightsum = sum(weight * flag for weight, flag
                                           in zip(uniq_pair_ballots_weights, cand2_vs_cand1))

            # add counts to df
            condorcet_count_df.loc[cand1, cand2] = cand1_vs_cand2_weightsum
            condorcet_count_df.loc[cand2, cand1] = cand2_vs_cand1_weightsum

            # calculate percent
            if sum_weighted_ballots:
                cand1_percent = (cand1_vs_cand2_weightsum / sum_weighted_ballots) * 100
                cand2_percent = (cand2_vs_cand1_weightsum / sum_weighted_ballots) * 100
            else:
                cand1_percent = 0
                cand2_percent = 0

            # add to df
            condorcet_percent_df.loc[cand1, cand2] = cand1_percent
            condorcet_percent_df.loc[cand2, cand1] = cand2_percent

        # find condorcet winner and set index name to include winner
        condorcet_winner = None

        if len(candidate_set) == 1:

            condorcet_winner = candidate_set[0]
        else:

            for cand in candidate_set:

                not_cand = set(candidate_set) - {cand}
                all_winner = all(condorcet_percent_df.loc[cand, not_cand] > 50)

                if all_winner:
                    if condorcet_winner is None:
                        condorcet_winner = cand
                    else:
                        print("cannottt be more than one condorcet winner!!!!")
                        exit(1)

        # convert decimal to float
        condorcet_count_df = condorcet_count_df.astype(float).round(3)
        condorcet_percent_df = condorcet_percent_df.astype(float).round(3)

        return condorcet_count_df, condorcet_percent_df, condorcet_winner

    def first_second_tables(self):
        """
        Return two pandas tables with candidates as columns and first row showing distribution of first round votes.
        Subsequent rows indicate second choice vote distribution for each column.

        first table is vote counts
        second table is percentages
        """

        candidate_set = BallotMarks.combine_writein_marks(self.get_candidates())
        candidate_set = sorted(candidate_set.unique_candidates)

        cleaned_dict = self.get_cvr_dict(self._contest_rule_set_name)
        ballot_set = [{'ranks': ranks.marks, 'weight': weight}
                      for ranks, weight in zip(cleaned_dict['ballot_marks'], cleaned_dict['weight'])]

        # create data frame that will be populated and output
        percent_no_exhaust_df = pd.DataFrame(util.NAN, index=['first_choice', *candidate_set], columns=candidate_set)
        percent_df = pd.DataFrame(util.NAN, index=['first_choice', *candidate_set, 'exhaust'], columns=candidate_set)
        count_df = pd.DataFrame(util.NAN, index=['first_choice', *candidate_set, 'exhaust'], columns=candidate_set)

        # group ballots by first choice
        first_choices = {cand: [] for cand in candidate_set}
        for b in ballot_set:
            if len(b['ranks']) >= 1:
                first_choices[b['ranks'][0]].append(b)

        # sum total first round votes
        total_first_round_votes = 0
        for cand in first_choices:
            total_first_round_votes += sum([b['weight'] for b in first_choices[cand]])

        # add first choices to tables
        # and calculate second choices
        for cand in candidate_set:

            ############################################################
            # update first round table values
            first_choice_count = sum([b['weight'] for b in first_choices[cand]])
            first_choice_percent = (first_choice_count / total_first_round_votes) * 100

            count_df.loc['first_choice', cand] = first_choice_count
            percent_df.loc['first_choice', cand] = first_choice_percent
            percent_no_exhaust_df.loc['first_choice', cand] = first_choice_percent

            ############################################################
            # calculate second choices, group second choices by candidate
            possible_second_choices = list(set(candidate_set) - {cand})
            second_choices = {backup_cand: [] for backup_cand in possible_second_choices + ['exhaust']}

            # group ballots by second choices
            for b in first_choices[cand]:
                if len(b['ranks']) >= 2:
                    second_choices[b['ranks'][1]].append(b['weight'])
                else:
                    second_choices['exhaust'].append(b['weight'])

            # sum total second round votes
            total_second_choices = 0
            total_second_choices_no_exhaust = 0
            for backup_cand in second_choices:
                total_second_choices += sum(second_choices[backup_cand])
                if backup_cand != 'exhaust':
                    total_second_choices_no_exhaust += sum(second_choices[backup_cand])

            # count second choices and add to table
            for backup_cand in second_choices:

                # fill in second choice values in table
                second_choice_count = sum(second_choices[backup_cand])

                # if there are not backup votes fill with zeros
                if total_second_choices == 0:
                    second_choice_percent = 0
                else:
                    second_choice_percent = (second_choice_count / total_second_choices) * 100

                if total_second_choices_no_exhaust == 0:
                    second_choice_percent_no_exhaust = 0
                else:
                    second_choice_percent_no_exhaust = (second_choice_count / total_second_choices_no_exhaust) * 100

                count_df.loc[backup_cand, cand] = second_choice_count
                percent_df.loc[backup_cand, cand] = second_choice_percent
                if backup_cand != 'exhaust':
                    percent_no_exhaust_df.loc[backup_cand, cand] = second_choice_percent_no_exhaust

        count_df = count_df.astype(float).round(3)
        percent_df = percent_df.astype(float).round(3)
        percent_no_exhaust_df = percent_no_exhaust_df.astype(float).round(3)

        return count_df, percent_df, percent_no_exhaust_df

    def cumulative_ranking_tables(self):
        """
        Return cumulative ranking tables. Rows are candidate names and columns are rank numbers.
        Reading across columns, the tables show the accumulating count/percentage of ballots that marked
        a candidate as more ranks are considered. The final column shows the count/percentage of ballots
        that never marked the candidate.
        """

        # get inputs
        candidate_set = BallotMarks.combine_writein_marks(self.get_candidates())
        candidate_set = sorted(candidate_set.unique_candidates)

        # ballot rank limit
        ballot_length = self.stats()[0]['rank_limit'].item()

        # get cleaned ballots
        cleaned_dict = self.get_cvr_dict(self._contest_rule_set_name)
        ballot_set = [{'ranks': ranks.marks + (['NA'] * (ballot_length - len(ranks.marks))), 'weight': weight}
                      for ranks, weight in zip(cleaned_dict['ballot_marks'], cleaned_dict['weight'])]

        # total ballots
        total_ballots = sum([d['weight'] for d in ballot_set])

        # create data frame that will be populated and output
        col_names = ["Rank " + str(i + 1) for i in range(ballot_length)] + ['Did Not Rank']
        cumulative_percent_df = pd.DataFrame(util.NAN, index=candidate_set, columns=col_names)
        cumulative_count_df = pd.DataFrame(util.NAN, index=candidate_set, columns=col_names)

        # tally candidate counts by rank
        rank_counts = []
        for rank in range(0, ballot_length):
            rank_cand_set = set([b['ranks'][rank] for b in ballot_set]) - {'NA'}
            current_rank_count = {cand: 0 for cand in rank_cand_set}
            for cand in rank_cand_set:
                current_rank_count[cand] = sum(b['weight'] for b in ballot_set if cand == b['ranks'][rank])
            rank_counts.append(current_rank_count)

        # accumulate ballot counts that rank candidates
        cumulative_counter = {cand: 0 for cand in candidate_set}
        for rank in range(0, ballot_length):
            for cand in candidate_set:
                # if candidate has any marks for this rank, accumulate them
                if cand in rank_counts[rank]:
                    cumulative_counter[cand] += rank_counts[rank][cand]
                # update tables
                cumulative_count_df.loc[cand, 'Rank ' + str(rank + 1)] = cumulative_counter[cand]
                cumulative_percent_df.loc[cand, 'Rank ' + str(rank + 1)] = cumulative_counter[cand] * 100 / total_ballots

        # fill in Did Not Rank column
        for cand in candidate_set:
            cumulative_count_df.loc[cand, 'Did Not Rank'] = total_ballots - cumulative_counter[cand]
            cumulative_percent_df.loc[cand, 'Did Not Rank'] = (total_ballots - cumulative_counter[cand]) * 100 / total_ballots

        cumulative_count_df = cumulative_count_df.astype(float).round(3)
        cumulative_percent_df = cumulative_percent_df.astype(float).round(3)

        return cumulative_count_df, cumulative_percent_df

    def winner_choice_position_distribution_table(self, tabulation_num: int = 1) -> Optional[pd.DataFrame]:

        winner = self._tabulation_winner(tabulation_num=tabulation_num)
        contest_cvr_dl = self.get_cvr_dict(rule_set_name=self._contest_rule_set_name)

        if len(winner) > 1:
            return None

        winner = winner[0]
        rank_limit = self.stats()[0]['rank_limit'].item()
        winner_final_round_count = self._final_round_winner_vote(tabulation_num=tabulation_num)
        final_weight_distrib = self.get_final_weight_distrib(tabulation_num=tabulation_num)

        winner_positions = {k: 0 for k in range(1, rank_limit+1)}
        for b, weight, final_distrib in zip(contest_cvr_dl['ballot_marks'], contest_cvr_dl['weight'], final_weight_distrib):

            if winner == final_distrib[-1][0]:
                winner_position = b.marks.index(winner)
                winner_positions[winner_position + 1] += weight

        if sum(winner_positions.values()) != winner_final_round_count:
            raise RuntimeError()

        winner_positions = {k: 100 * v / winner_final_round_count for k, v in winner_positions.items()}

        choice_position_df = self.stats()[0][[
            'jurisdiction',
            'state',
            'date',
            'year',
            'office',
            'unique_id',
            'winner',
            'rank_limit',
            'n_candidates',
            'n_rounds'
            ]]

        for position, percent in sorted(winner_positions.items()):
            choice_position_df[f'choice{position}'] = float(round(percent, 2))

        return choice_position_df

    def first_choice_to_finalist_table(self, tabulation_num: int = 1) -> pd.DataFrame:

        df = None

        # who had any ballot weight allotted
        finalist_candidates = list(self.finalist_candidates(tabulation_num=tabulation_num)) + ['exhaust']
        candidate_set = sorted(self._contest_candidates.unique_candidates)

        ballot_set = [{'ranks': ranks, 'weight': weight, 'weight_distrib': distrib}
                      for ranks, weight, distrib
                      in zip(self.get_initial_ranks(tabulation_num=tabulation_num),
                             self.get_initial_weights(tabulation_num=tabulation_num),
                             self.get_final_weight_distrib(tabulation_num=tabulation_num))]

        index_label = "Ballots with first choice:"
        n_ballots_label = "Number of Ballots"

        colname_dict = {cand: "% of votes to " + cand for cand in finalist_candidates}

        rows = candidate_set
        cols = [n_ballots_label] + list(colname_dict.values())
        df = pd.DataFrame(index=rows, columns=cols + ['percent_sum'])
        df.index.name = index_label

        # group ballots by first choice
        first_choices = {cand: [] for cand in candidate_set}
        for b in ballot_set:
            if len(b['ranks']) >= 1 and b['ranks'][0] in first_choices:
                first_choices[b['ranks'][0]].append(b)

        for cand in candidate_set:

            total_first_choice_ballots = sum(b['weight'] for b in first_choices[cand])
            df.loc[cand, n_ballots_label] = total_first_choice_ballots

            if total_first_choice_ballots:

                redistrib = {opponent: 0 for opponent in finalist_candidates}
                for b in first_choices[cand]:
                    for el in b['weight_distrib']:
                        if el[0] == 'empty':
                            redistrib['exhaust'] += el[1]
                        else:
                            redistrib[el[0]] += el[1]

                redistrib_total_check = 0
                for opponent in redistrib:
                    redistrib_percent = redistrib[opponent] / total_first_choice_ballots * 100
                    df.loc[cand, colname_dict[opponent]] = redistrib_percent
                    redistrib_total_check += redistrib_percent
                df.loc[cand, 'percent_sum'] = redistrib_total_check

            else:
                for opponent in finalist_candidates:
                    df.loc[cand, colname_dict[opponent]] = 0
                df.loc[cand, 'percent_sum'] = 0

        df = df.astype(float).round(3)
        return df

    def round_by_round_table(self, tabulation_num: int = 1) -> pd.DataFrame:

        num_rounds = self.n_rounds(tabulation_num=tabulation_num)

        first_round_exhaust = self.stats()[tabulation_num-1]['total_pretally_exhausted'].item()

        # get rcv results
        first_round_dict = self.get_round_tally_dict(1, tabulation_num=tabulation_num)
        rounds_full = [self.get_round_tally_tuple(i, tabulation_num=tabulation_num) for i in range(1, num_rounds + 1)]
        transfers = [self.get_round_transfer_dict(i, tabulation_num=tabulation_num) for i in range(1, num_rounds + 1)]

        # reformat contest outputs into useful dicts
        cand_outcomes = self.get_candidate_outcomes(tabulation_num=tabulation_num)

        # reorder candidate names
        # winners in ascending order of round won
        # followed by losers in descending order of round lost
        reorder_dicts = []
        for d in cand_outcomes:

            if d['round_elected']:
                d['order'] = -1 * (1 / d['round_elected'])
            else:
                d['order'] = 1 / d['round_eliminated']

            reorder_dicts.append(d)

        ordered_candidates_names = [d['name']
                                    for d in sorted(reorder_dicts,
                                                    key=lambda x: (x['order'],
                                                                   -first_round_dict[x['name']],
                                                                   x['name']))
                                    ]

        # setup data frame
        row_names = ordered_candidates_names + ['exhaust']
        rcv_df = pd.DataFrame(NAN, index=row_names + ['colsum'], columns=['candidate'])
        rcv_df.loc[row_names + ['colsum'], 'candidate'] = row_names + ['colsum']

        # loop through rounds
        for rnd in range(1, num_rounds + 1):

            rnd_info = {rnd_cand: rnd_tally for rnd_cand, rnd_tally in zip(*rounds_full[rnd-1])}
            rnd_info['exhaust'] = 0

            rnd_transfer = dict(transfers[rnd-1])

            # add round data
            for cand in row_names:

                rnd_percent_col = 'r' + str(rnd) + '_active_percent'
                rnd_count_col = 'r' + str(rnd) + '_count'
                rnd_transfer_col = 'r' + str(rnd) + '_transfer'

                rcv_df.loc[cand, rnd_percent_col] = 100*(rnd_info[cand]/sum(rnd_info.values()))
                rcv_df.loc[cand, rnd_count_col] = rnd_info[cand]
                rcv_df.loc[cand, rnd_transfer_col] = rnd_transfer[cand]

            # maintain cumulative exhaust total
            if rnd == 1:
                rcv_df.loc['exhaust', rnd_count_col] = first_round_exhaust
            else:
                last_rnd_count_col = 'r' + str(rnd-1) + '_count'
                last_rnd_transfer_col = 'r' + str(rnd-1) + '_transfer'
                current_rnd_count_val = sum(rcv_df.loc['exhaust', [last_rnd_count_col, last_rnd_transfer_col]].astype(float))
                rcv_df.loc['exhaust', rnd_count_col] = current_rnd_count_val

            # sum round columns
            rcv_df.loc['colsum', rnd_count_col] = sum(rcv_df.loc[row_names, rnd_count_col].astype(float))
            rcv_df.loc['colsum', rnd_transfer_col] = sum(rcv_df.loc[row_names, rnd_transfer_col].astype(float))
            rcv_df.loc['colsum', rnd_percent_col] = sum(rcv_df.loc[row_names, rnd_percent_col].astype(float))

        # # convert from decimal to float
        rcv_df.loc[row_names + ['colsum'], rcv_df.columns != "candidate"] = \
            rcv_df.loc[row_names + ['colsum'], rcv_df.columns != "candidate"].astype(float).round(3)

        # remove rownames
        rcv_df = rcv_df.reset_index(drop=True)
        return rcv_df

    def annotated_cvr_table(self):
        dfs = [super().annotated_cvr_table(), self._contest_stat_table]
        return pd.concat(dfs, axis='columns', sort=False)

    # def outcome_cvr_table(self):

    #     dfs = []
    #     for iTab in range(1, self.n_tabulations()+1):

    #         # get cvr df
    #         cvr = ballots.cvr_table(rcv_obj.ctx, table_format=cvr_format)
    #         cvr['ballot_split_ID'] = cvr.index + 1
    #         cvr['exhaustion_check'] = rcv_obj.exhaustion_check(tabulation_num=iTab)

    #         # duplicate ballot rows in df for each time the ballot was split
    #         final_weight_distrib = rcv_obj.get_final_weight_distrib(tabulation_num=iTab)

    #         # convert final weights to string
    #         str_convert = []
    #         for tl in final_weight_distrib:
    #             str_convert.append(";".join([pair[0] + ":" + str(float(pair[1])) for pair in tl]))

    #         # duplicate rows
    #         cvr['final_round_allocation'] = str_convert
    #         split_allocation = cvr['final_round_allocation'].str.split(';').apply(pd.Series, 1).stack()
    #         split_allocation.index = split_allocation.index.droplevel(-1)
    #         split_allocation.name = 'final_round_allocation'
    #         del cvr['final_round_allocation']
    #         cvr = cvr.join(split_allocation)

    #         # split allocation column out to candidate and weight
    #         cvr[['final_allocation', 'weight']] = cvr.final_round_allocation.str.split(":", expand=True)

    #         # create allocation column that marks all inactive ballots as 'inactive' and
    #         # inactive_type column that marks all candidates as 'NA' but specifies the kind of inactive ballot
    #         # exhausted, undervote, etc
    #         cvr['inactive_type'] = cvr['final_allocation']

    #         # inactive_type column
    #         cvr.loc[cvr['inactive_type'] != 'empty', 'inactive_type'] = 'NA'
    #         cvr.loc[cvr['inactive_type'] == 'empty', 'inactive_type'] = cvr.loc[cvr['inactive_type'] == 'empty', 'exhaustion_check']

    #         # final_allocation_column
    #         cvr['final_allocation'] = cvr['final_allocation'].replace(to_replace="empty", value="inactive")

    #         # remove intermediate columns
    #         del cvr['final_round_allocation']
    #         del cvr['exhaustion_check']

    #         # reorder columns
    #         ballot_split_ID_col = cvr.pop('ballot_split_ID')
    #         cvr.insert(0, 'ballot_split_ID', ballot_split_ID_col)

    #         uid = rcv_obj.file_stub(tabulation_num=iTab)
    #         outfile = cvr_allocation_dir / f'{uid}_ballot_allocation.csv'
    #         cvr.to_csv(util.longname(outfile), index=False)



#     cvr_allocation_dir = results_dir / 'cvr_ballot_allocation'
#     util.verifyDir(cvr_allocation_dir)

#     if cvr_format == 'rank':
#         cvr_allocation_dir = cvr_allocation_dir / "rank_format"
#     elif cvr_format == 'candidate':
#         cvr_allocation_dir = cvr_allocation_dir / "candidate_format"
#     util.verifyDir(cvr_allocation_dir)

#

    def write_round_by_round_json(self, save_dir: Union[pathlib.Path, str]) -> None:

        n_tabulations = self.n_tabulations()
        for iTab in range(1, n_tabulations+1):

            # get rcv results
            n_rounds = self.n_rounds(tabulation_num=iTab)
            tally_tuples = [self.get_round_tally_tuple(round_num=i, tabulation_num=iTab)
                            for i in range(1, n_rounds+1)]
            transfer_dicts = [self.get_round_transfer_dict(round_num=i, tabulation_num=iTab)
                            for i in range(1, n_rounds+1)]

            outcomes = self.get_candidate_outcomes(tabulation_num=iTab)

            json_dict = {'config': {'notes': self.stats()[iTab-1]['notes'].item(),
                                    'date': self.stats()[iTab-1]['date'].item(),
                                    'jurisdiction': self.stats()[iTab-1]['jurisdiction'].item(),
                                    'office': self.stats()[iTab-1]['office'].item(),
                                    'threshold': '50%'},
                        'results': []}

            for i in range(0, n_rounds):

                round_num = i + 1
                tally_dict = {cand: str(int(tally)) for cand, tally in zip(*tally_tuples[i]) if tally > 0}
                transfer_list = []

                # who had an outcomes this round
                elected = [d for d in outcomes if d['round_elected'] == round_num]
                eliminated = [d for d in outcomes if d['round_eliminated'] == round_num]

                for d in elected:
                    transfer_list.append({'elected': d['name'], 'transfers': {}})

                for d in eliminated:
                    if i == n_rounds-1:  # no final round transfer
                        transfer_list.append({'eliminated': d['name'], 'transfers': {}})
                    elif transfer_dicts[i][d['name']] == 0:  # zero vote elimination, no votes transferred
                        transfer_list.append({'eliminated': d['name'], 'transfers': {}})
                    else:  # loser transfer
                        # remove negative transfer
                        round_transfer = {key: str(int(val)) for key, val in transfer_dicts[i].items() if val > 0}
                        if 'exhaust' in round_transfer:  # small rename
                            round_transfer['exhausted'] = round_transfer['exhaust']
                            del round_transfer['exhaust']
                        transfer_list.append({'eliminated': d['name'], 'transfers': round_transfer})

                json_dict['results'].append(
                    {'round': round_num,
                     'tally': tally_dict,
                     'tallyResults': transfer_list})

            uid = self.stats()[iTab-1]['unique_id'].item()

            save_path = pathlib.Path(save_dir) / 'round_by_round_json'
            save_path.mkdir(exist_ok=True)

            outfile = open(save_path / f"{uid}_tab{iTab}.json", 'w')
            json.dump(json_dict, outfile)
            outfile.close()

    def candidate_rank_usage_table(self):

        # set up vars
        contest_candidates = self._contest_candidates.unique_candidates
        rank_limit = self._summary_cvr_stat_table['rank_limit'].item()
        candidate_outcomes = {dikt['name']: dikt for dikt in self.get_candidate_outcomes(tabulation_num=1)}
        first_round_dict = self.get_round_tally_dict(round_num=1, tabulation_num=1)
        first_round_leader = sorted(first_round_dict.items(), key=lambda x: -x[1])[0][0]

        contest_cvr_dl = self.get_cvr_dict(self._contest_rule_set_name)
        contest_cvr_ld = [{'ballot_marks': bm, 'weight': weight}
                          for bm, weight in zip(contest_cvr_dl['ballot_marks'], contest_cvr_dl['weight'])]

        # set up dataframe
        df_columns = [
            'contestID',
            'rank_limit',
            'n_rounds',
            'rcv_type',
            'winner',
            'first_round_percent',
            'first_round_count',
            'round_elected',
            'round_eliminated'
        ] + [
            f'ranked_{i}' for i in range(1, rank_limit+1)
        ] + [
            'ranked_2_or_more',
            'ranked_3_or_more',
            'mean_ranked',
            'all_candidate_mean_ranked',
            'non_leader_mean_ranked',
            'non_leader_no_writein_mean_ranked'
        ]
        df = pd.DataFrame(None, index=contest_candidates, columns=df_columns)

        # fill precomputed columns
        df['contestID'] = self._id_df['unique_id'].item()
        df['rank_limit'] = rank_limit
        df['n_rounds'] = self._summary_contest_stat_tables[0]['n_rounds'].item()
        df['rcv_type'] = self._summary_contest_stat_tables[0]['rcv_type'].item()
        df['winner'] = [True if candidate in self._tabulation_winner(tabulation_num=1) else False
                        for candidate in df.index]
        df['round_elected'] = [candidate_outcomes[candidate]['round_elected'] for candidate in df.index]
        df['round_eliminated'] = [candidate_outcomes[candidate]['round_eliminated'] for candidate in df.index]
        df['first_round_percent'] = [100 * first_round_dict[candidate] / sum(first_round_dict.values())
                                     for candidate in df.index]
        df['first_round_count'] = [first_round_dict[candidate] for candidate in df.index]

        # count rank usage by candidate
        rank_usage_count = {candidate: {i: 0 for i in range(1, rank_limit+1)} for candidate in contest_candidates}
        for ballot in contest_cvr_ld:
            if ballot['ballot_marks'].marks:
                first_candidate = ballot['ballot_marks'].marks[0]
                ranks_used = len(ballot['ballot_marks'].marks)
                rank_usage_count[first_candidate][ranks_used] += ballot['weight']

        # convert counts to percent and add to table
        all_candidate_rank_usage = {i: 0 for i in range(1, rank_limit+1)}
        non_leader_rank_usage = {i: 0 for i in range(1, rank_limit+1)}
        non_leader_no_writein_rank_usage = {i: 0 for i in range(1, rank_limit+1)}
        for candidate, candidate_rank_usage in rank_usage_count.items():

            candidate_total_ballots = sum(candidate_rank_usage.values())
            for ranks_used, ranks_used_count in candidate_rank_usage.items():

                all_candidate_rank_usage[ranks_used] += ranks_used_count

                if candidate != first_round_leader:
                    non_leader_rank_usage[ranks_used] += ranks_used_count

                if candidate != BallotMarks.WRITEIN:
                    non_leader_no_writein_rank_usage[ranks_used] += ranks_used_count

                df.loc[candidate, f'ranked_{ranks_used}'] = 0
                if candidate_total_ballots:
                    df.loc[candidate, f'ranked_{ranks_used}'] = 100 * ranks_used_count / candidate_total_ballots

            df.loc[candidate, 'ranked_2_or_more'] = 0
            df.loc[candidate, 'ranked_3_or_more'] = 0
            if candidate_total_ballots:
                ranked_2_or_more = sum(count for ranks_used, count in candidate_rank_usage.items() if ranks_used > 1)
                ranked_3_or_more = sum(count for ranks_used, count in candidate_rank_usage.items() if ranks_used > 2)
                df.loc[candidate, 'ranked_2_or_more'] = 100 * ranked_2_or_more / candidate_total_ballots
                df.loc[candidate, 'ranked_3_or_more'] = 100 * ranked_3_or_more / candidate_total_ballots

            df.loc[candidate, 'mean_ranked'] = 0
            if candidate_total_ballots:
                weighted_sum = sum(count * ranks_used for ranks_used, count in candidate_rank_usage.items())
                df.loc[candidate, 'mean_ranked'] = weighted_sum / candidate_total_ballots

        all_candidate_weighted_sum = sum(count * ranks_used for ranks_used, count in all_candidate_rank_usage.items())
        non_leader_weighted_sum = sum(count * ranks_used for ranks_used, count in non_leader_rank_usage.items())
        non_leader_no_writein_weighted_sum = sum(count * ranks_used
                                                 for ranks_used, count in non_leader_no_writein_rank_usage.items())

        if sum(all_candidate_rank_usage.values()):
            df['all_candidate_mean_ranked'] = all_candidate_weighted_sum / sum(all_candidate_rank_usage.values())
        if sum(non_leader_rank_usage.values()):
            df['non_leader_mean_ranked'] = non_leader_weighted_sum / sum(non_leader_rank_usage.values())
        if sum(non_leader_no_writein_rank_usage.values()):
            df['non_leader_no_writein_mean_ranked'] = non_leader_no_writein_weighted_sum / sum(non_leader_no_writein_rank_usage.values())

        return df




# import os

# import pandas as pd

# import rcv_cruncher.ballots as ballots
# import rcv_cruncher.misc_tabulation as misc_tabulation
# import rcv_cruncher.util as util


# def write_converted_cvr(contest, results_dir, cvr_format='rank'):
#     """
#     Convert cvr into common csv format and write out
#     """
#     uid = contest['uid'] if not contest.get('split_id') else contest.get('split_id')

#     if cvr_format == 'rank':
#         format_dir = results_dir / "rank_format"
#     elif cvr_format == 'candidate':
#         format_dir = results_dir / "candidate_format"
#     util.verifyDir(format_dir)

#     outfile = format_dir / f"{uid}.csv"

#     cvr = ballots.cvr_table(contest, table_format=cvr_format)
#     cvr.to_csv(util.longname(outfile), index=False)


# def write_converted_cvr_annotated(rcv_obj, results_dir, cvr_format='candidate'):
#     """
#     Convert cvr into common csv format, with appended columns and write out
#     """
#     cvr_allocation_dir = results_dir / 'cvr_ballot_allocation'
#     util.verifyDir(cvr_allocation_dir)

#     if cvr_format == 'rank':
#         cvr_allocation_dir = cvr_allocation_dir / "rank_format"
#     elif cvr_format == 'candidate':
#         cvr_allocation_dir = cvr_allocation_dir / "candidate_format"
#     util.verifyDir(cvr_allocation_dir)

#     for iTab in range(1, rcv_obj.n_tabulations()+1):

#         # get cvr df
#         cvr = ballots.cvr_table(rcv_obj.ctx, table_format=cvr_format)
#         cvr['ballot_split_ID'] = cvr.index + 1
#         cvr['exhaustion_check'] = rcv_obj.exhaustion_check(tabulation_num=iTab)

#         # duplicate ballot rows in df for each time the ballot was split
#         final_weight_distrib = rcv_obj.get_final_weight_distrib(tabulation_num=iTab)

#         # convert final weights to string
#         str_convert = []
#         for tl in final_weight_distrib:
#             str_convert.append(";".join([pair[0] + ":" + str(float(pair[1])) for pair in tl]))

#         # duplicate rows
#         cvr['final_round_allocation'] = str_convert
#         split_allocation = cvr['final_round_allocation'].str.split(';').apply(pd.Series, 1).stack()
#         split_allocation.index = split_allocation.index.droplevel(-1)
#         split_allocation.name = 'final_round_allocation'
#         del cvr['final_round_allocation']
#         cvr = cvr.join(split_allocation)

#         # split allocation column out to candidate and weight
#         cvr[['final_allocation', 'weight']] = cvr.final_round_allocation.str.split(":", expand=True)

#         # create allocation column that marks all inactive ballots as 'inactive' and
#         # inactive_type column that marks all candidates as 'NA' but specifies the kind of inactive ballot
#         # exhausted, undervote, etc
#         cvr['inactive_type'] = cvr['final_allocation']

#         # inactive_type column
#         cvr.loc[cvr['inactive_type'] != 'empty', 'inactive_type'] = 'NA'
#         cvr.loc[cvr['inactive_type'] == 'empty', 'inactive_type'] = cvr.loc[cvr['inactive_type'] == 'empty', 'exhaustion_check']

#         # final_allocation_column
#         cvr['final_allocation'] = cvr['final_allocation'].replace(to_replace="empty", value="inactive")

#         # remove intermediate columns
#         del cvr['final_round_allocation']
#         del cvr['exhaustion_check']

#         # reorder columns
#         ballot_split_ID_col = cvr.pop('ballot_split_ID')
#         cvr.insert(0, 'ballot_split_ID', ballot_split_ID_col)

#         uid = rcv_obj.file_stub(tabulation_num=iTab)
#         outfile = cvr_allocation_dir / f'{uid}_ballot_allocation.csv'
#         cvr.to_csv(util.longname(outfile), index=False)


# def write_condorcet_tables(contest, results_dir):
#     """
#     Calculate and write condorcet tables (both count and percents) for contest
#     """
#     counts, percents, condorcet_winner = misc_tabulation.condorcet_tables(contest)

#     if condorcet_winner:
#         condorcet_str = "condorcet winner: " + condorcet_winner
#     else:
#         condorcet_str = "condorcet winner: None"

#     counts.index.name = condorcet_str
#     percents.index.name = condorcet_str

#     condorcet_table_dir = results_dir / 'condorcet'
#     util.verifyDir(condorcet_table_dir)

#     uid = contest['uid'] if not contest.get('split_id') else contest.get('split_id')
#     counts.to_csv(util.longname(condorcet_table_dir / f"{uid}_condorcet_count.csv"), float_format="%.2f")
#     percents.to_csv(util.longname(condorcet_table_dir / f"{uid}_condorcet_percent.csv"), float_format="%.2f")


# def write_first_second_tables(contest, results_dir):
#     """
#     Calculate and write first choice - second choice tables (both count and percents) for contest
#     """
#     counts, percents, percents_no_exhaust = misc_tabulation.first_second_tables(contest)

#     first_second_table_dir = results_dir / 'first_second'
#     util.verifyDir(first_second_table_dir)

#     uid = contest['uid'] if not contest.get('split_id') else contest.get('split_id')
#     counts.to_csv(util.longname(first_second_table_dir / f"{uid}_first_second_choices_count.csv"), float_format="%.2f")
#     percents.to_csv(util.longname(first_second_table_dir / f"{uid}_first_second_choices_percent.csv"), float_format="%.2f")
#     percents_no_exhaust.to_csv(
#       util.longname(first_second_table_dir / f"{uid}_first_second_choices_percent_no_exhaust.csv"), float_format="%.2f")

# def write_cumulative_ranking_tables(contest, results_dir):
#     """
#     Calculate and write cumulative ranking tables (both count and percent) for contest
#     """
#     counts, percents = misc_tabulation.cumulative_ranking_tables(contest)

#     cumulative_ranking_dir = results_dir / 'cumulative_ranking'
#     util.verifyDir(cumulative_ranking_dir)

#     uid = contest['uid'] if not contest.get('split_id') else contest.get('split_id')
#     counts.to_csv(util.longname(cumulative_ranking_dir / f"{uid}_cumulative_ranking_count.csv"), float_format="%.2f")
#     percents.to_csv(util.longname(cumulative_ranking_dir / f"{uid}_cumulative_ranking_percent.csv"), float_format="%.2f")


# def write_rank_usage_tables(contest, results_dir):
#     """
#     Calculate and write rank usage statistics.
#     """
#     df = misc_tabulation.rank_usage_tables(contest)

#     rank_usage_table_dir = results_dir / 'rank_usage'
#     util.verifyDir(rank_usage_table_dir)

#     uid = contest['uid'] if not contest.get('split_id') else contest.get('split_id')
#     df.to_csv(util.longname(rank_usage_table_dir / f'{uid}_rank_usage.csv'), float_format='%.2f')


# def write_opponent_crossover_tables(contest, results_dir):
#     """
#     Calculate crossover support between candidates and write table.
#     """
#     count_df, percent_df = misc_tabulation.crossover_table(contest)

#     opponent_crossover_table_dir = results_dir / 'crossover_support'
#     util.verifyDir(opponent_crossover_table_dir)

#     uid = contest['uid'] if not contest.get('split_id') else contest.get('split_id')
#     count_df.to_csv(util.longname(opponent_crossover_table_dir / f'{uid}_crossover_support_count.csv'), float_format='%.2f')
#     percent_df.to_csv(util.longname(opponent_crossover_table_dir / f'{uid}_crossover_support_percent.csv'), float_format='%.2f')


# def write_first_to_finalist_tables(rcv_obj, results_dir):
#     """
#     Calculate distribution of ballots allocated to non-finalists during first round and their transfer
#     to eventual finalists.
#     """

#     first_to_finalist_table_dir = results_dir / 'first_choice_to_finalist'
#     util.verifyDir(first_to_finalist_table_dir)

#     dfs = misc_tabulation.first_choice_to_finalist_table(rcv_obj)

#     for iTab in range(1, len(dfs) + 1):
#         uid = rcv_obj.file_stub(tabulation_num=iTab)
#         csv_fname = first_to_finalist_table_dir / f'{uid}_first_to_finalist.csv'
#         dfs[iTab-1].to_csv(util.longname(csv_fname), float_format='%.2f')


# def prepare_candidate_details(obj):
#     """
#     Return pandas data frame of candidate info with round-by-round vote counts
#     """
#     all_dfs = []

#     for iTab in range(1, obj.n_tabulations()+1):

#         uid = obj.file_stub(tabulation_num=iTab)
#         n_rounds = obj.n_rounds(tabulation_num=iTab)

#         # get rcv results
#         rounds_full = [obj.get_round_tally_tuple(i, tabulation_num=iTab) for i in range(1, n_rounds + 1)]
#         cand_outcomes = obj.get_candidate_outcomes(tabulation_num=iTab)

#         # reformat contest outputs into useful dicts
#         cand_outcomes_dict = {d['name']: d for d in cand_outcomes}
#         rounds_full_dict = [{cand: float(count) for cand, count in zip(*rnd_count)}
#                             for rnd_count in rounds_full]

#         # reorder candidate names
#         # winners in ascending order of round won
#         # followed by losers in descending order of round lost
#         reorder_dicts = []
#         for d in cand_outcomes:

#             # don't add candidates if they received zero votes throughout the contest.
#             if sum(rounds_dict[d['name']] for rounds_dict in rounds_full_dict) == 0:
#                 continue

#             if d['round_elected']:
#                 d['order'] = -1 * (1/d['round_elected'])
#             else:
#                 d['order'] = 1/d['round_eliminated']

#             reorder_dicts.append(d)

#         ordered_candidates_names = [d['name'] for d in sorted(reorder_dicts, key=lambda x: x['order'])]

#         # create table
#         colnames = ['unique_id', 'candidate', 'round_elected', 'round_eliminated'] + \
#                    ['round_' + str(i) + '_vote' for i in range(1, n_rounds + 1)]

#         # assemble rows
#         cand_rows = []
#         for cand in ordered_candidates_names:

#             cand_rows.append([uid,
#                               cand,
#                               cand_outcomes_dict[cand]['round_elected'],
#                               cand_outcomes_dict[cand]['round_eliminated']] +
#                              [d[cand] for d in rounds_full_dict])

#         df = pd.DataFrame(cand_rows, columns=colnames)
#         all_dfs.append(df)

#     return all_dfs


# def write_rcv_rounds(obj, results_dir):
#     """
#     Write out rcv contest round-by-round counts and transfers
#     """

#     round_by_round_dir = results_dir / 'round_by_round'
#     util.verifyDir(round_by_round_dir)

#     rcv_dfs = misc_tabulation.round_by_round(obj)

#     for iTab, rcv_df in enumerate(rcv_dfs, start=1):
#         uid = obj.file_stub(tabulation_num=iTab)
#         rcv_df.to_csv(util.longname(round_by_round_dir / f'{uid}_round_by_round.csv'), index=False)


# def write_ballot_debug_info(obj, results_dir):
#     """
#     Write ballot debug CVR
#     """
#     for iTab in range(1, obj.n_tabulations()+1):

#         ballot_debug_df = obj.ballot_debug_df(tabulation_num=iTab)

#         output_dir = results_dir / 'ballot_stats_debug'
#         util.verifyDir(output_dir)

#         uid = obj.file_stub(tabulation_num=iTab)
#         ballot_debug_df.to_csv(util.longname(output_dir / f'{uid}_ballot_debug.csv'), index=False)



from __future__ import annotations
from typing import (Dict, Tuple, Type, Union, List, Optional)

import abc
import collections
import decimal
import random
import pathlib

import pandas as pd

import rcv_cruncher.util as util

from rcv_cruncher.cvr.base import CastVoteRecord
from rcv_cruncher.marks import BallotMarks
from rcv_cruncher.rcv.stats import RCV_stats
from rcv_cruncher.rcv.tables import RCV_tables


class RCV(abc.ABC, CastVoteRecord, RCV_stats, RCV_tables):

    @staticmethod
    def get_variant_group(rcv_obj: Type[RCV]) -> str:
        return 'single_winner' if len(rcv_obj._all_winners()) == 1 else 'multi_winner'

    @staticmethod
    def get_variant_name(rcv_obj: Type[RCV]) -> str:
        return rcv_obj.__class__.__name__

    @staticmethod
    def get_winner_choice_position_distribution_table(rcv_obj: Type[RCV],
                                                      tabulation_num: int = 1) -> Optional[pd.DataFrame]:
        return rcv_obj.winner_choice_position_distribution_table(tabulation_num=tabulation_num)

    @staticmethod
    def get_first_choice_to_finalist_table(rcv_obj: Type[RCV], tabulation_num: int = 1) -> pd.DataFrame:
        return rcv_obj.first_choice_to_finalist_table(tabulation_num=tabulation_num)

    @staticmethod
    def get_candidate_rank_usage_table(rcv_obj: Type[CastVoteRecord]) -> pd.DataFrame:
        return rcv_obj.candidate_rank_usage_table()

    @staticmethod
    def write_first_choice_to_finalist_table(rcv_obj: Type[RCV],
                                             save_dir: Union[str, pathlib.Path] = None) -> None:

        save_path = pathlib.Path(save_dir) / 'first_choice_to_finalist'
        save_path.mkdir(exist_ok=True)

        uid = rcv_obj.stats()[0]['unique_id'].item()
        for iTab in range(1, rcv_obj.n_tabulations() + 1):
            df = rcv_obj.first_choice_to_finalist_table(tabulation_num=iTab)
            df.to_csv(save_path / f'{uid}_tab{iTab}.csv')

    @staticmethod
    def get_round_by_round_table(rcv_obj: Type[RCV], tabulation_num: int = 1) -> pd.DataFrame:
        return rcv_obj.round_by_round_table(tabulation_num=tabulation_num)

    @staticmethod
    def write_round_by_round_table(rcv_obj: Type[RCV],
                                   save_dir: Union[str, pathlib.Path] = None) -> None:

        save_path = pathlib.Path(save_dir) / 'round_by_round_table'
        save_path.mkdir(exist_ok=True)

        uid = rcv_obj.stats()[0]['unique_id'].item()
        for iTab in range(1, rcv_obj.n_tabulations() + 1):
            df = rcv_obj.round_by_round_table(tabulation_num=iTab)
            df.to_csv(save_path / f'{uid}_tab{iTab}.csv', index=False)

    # override me
    @abc.abstractmethod
    def _set_round_winners(self) -> None:
        """
        This function should set self.round_winners to the list of candidates that won the round
        """
        pass

    # override me
    @abc.abstractmethod
    def _contest_not_complete(self) -> bool:
        """
        This function should return True if another round should be evaluated and False
        is the contest should complete.
        """
        pass

    # override me
    @abc.abstractmethod
    def _calc_round_transfer(self) -> None:
        """
        This function should append a dictionary to self._tabulations[self._tab_num-1]['transfers'] containing:
        candidate names as keys, plus one key for 'exhaust' and any other keys for transfer categories
        values as round transfer flows.
        """
        pass

    # override me
    def _win_threshold(self) -> Optional[Union[int, float]]:
        """
        This function should return the win threshold used in the contest
        OR return 'dynamic' if threshold changes with each round.
        """
        return None

    # override me, if ballots should be split/re-weighted prior to next round
    # such as in fractional transfer contests
    def _update_weights(self) -> None:
        pass

    # override me, if you need to do multiple iterations of rcv, e.x. utah sequential rcv
    def _run_contest(self) -> None:
        # run tabulation
        self._new_tabulation()
        self._tabulate()

    def __init__(self,
                 exhaust_on_duplicate_candidate_marks: bool = False,
                 exhaust_on_overvote_marks: bool = False,
                 exhaust_on_repeated_skipped_marks: bool = False,
                 treat_combined_writeins_as_exhaustable_duplicates: bool = True,
                 combine_writein_marks: bool = True,
                 exclude_writein_marks: bool = False,
                 n_winners: Optional[int] = None,
                 multi_winner_rounds: Optional[bool] = None,
                 bottoms_up_threshold: Optional[float] = None,
                 *args, **kwargs) -> None:

        # INIT CVR
        super().__init__(*args, **kwargs)

        # APPLY CONTEST RULES
        self._contest_rule_set_name = '__contest'
        self.add_rule_set(self._contest_rule_set_name,
                          BallotMarks.new_rule_set(
                              combine_writein_marks=combine_writein_marks,
                              exclude_writein_marks=exclude_writein_marks,
                              exclude_duplicate_candidate_marks=True,
                              exclude_overvote_marks=True,
                              exclude_skipped_marks=True,
                              treat_combined_writeins_as_exhaustable_duplicates=treat_combined_writeins_as_exhaustable_duplicates,
                              exhaust_on_duplicate_candidate_marks=exhaust_on_duplicate_candidate_marks,
                              exhaust_on_overvote_marks=exhaust_on_overvote_marks,
                              exhaust_on_repeated_skipped_marks=exhaust_on_repeated_skipped_marks
                          ))

        # CONTEST INPUTS
        self._bottoms_up_threshold = None
        if bottoms_up_threshold is not None:
            self._bottoms_up_threshold = decimal.Decimal(str(bottoms_up_threshold))
        self._n_winners = n_winners
        self._multi_winner_rounds = multi_winner_rounds
        self._contest_candidates = self.get_candidates(self._contest_rule_set_name)
        self._contest_cvr_ld = None
        self._reset_ballots()

        # INIT STATE INFO

        # contest-level
        self._tab_num = 0
        self._tabulations = []

        # tabulation-level
        self._inactive_candidates = []
        self._removed_candidates = []

        # round-level
        self._round_num = 0
        self._round_winners = []
        self._round_loser = None

        # RUN
        self._run_contest()

        # CONTEST STATS
        self._contest_stat_table = None
        self._compute_contest_stat_table()

        self._summary_contest_stat_tables = None
        self._compute_summary_contest_stat_tables()

        self._summary_contest_split_stat_tables = None

    def stats(self,
              keep_decimal_type: bool = False,
              add_split_stats: bool = False,
              add_id_info: bool = True) -> List[pd.DataFrame]:

        # start with cvr stats, 1 set per cvr
        cvr_stats = self._summary_cvr_stat_table.copy()

        # add on the contest stats for each tabulation
        contest_stats = [pd.concat([cvr_stats, df], axis='columns', sort=False)
                         for df in self._summary_contest_stat_tables]

        # add on the id info
        if add_id_info:
            contest_stats = [pd.concat([self._id_df, df], axis='columns', sort=False)
                             for df in contest_stats]

        if add_split_stats:

            self._make_split_filter_dict()
            self._compute_summary_cvr_split_stat_table()
            self._compute_summary_contest_split_stat_tables()
            cvr_split_stat_table = self._summary_cvr_split_stat_table
            contest_split_stat_tables = self._summary_contest_split_stat_tables

            if cvr_split_stat_table is not None and contest_split_stat_tables is not None:

                new_contest_stats = []
                for stat_table, split_stat_table in zip(contest_stats, contest_split_stat_tables):

                    # merge cvr split stats (1 per cvr) with current tabulation split stats
                    merged = cvr_split_stat_table.merge(split_stat_table, on=['split_field', 'split_value', 'split_id'])

                    # add in non split stat column generated in previous sections
                    merged = merged.assign(**{col: stat_table.at[0, col]
                                              if not isinstance(stat_table.at[0, col], tuple) else stat_table.at[0, col][0]
                                              for col in stat_table.columns})
                    merged = merged[
                        merged.columns.tolist()[-1 * len(stat_table.columns):] +
                        merged.columns.tolist()[:-1 * len(stat_table.columns)]
                    ]

                    new_contest_stats.append(merged)

                contest_stats = new_contest_stats

        if not keep_decimal_type:
            contest_stats = [t.applymap(util.decimal2float) for t in contest_stats]

        return contest_stats

    def _reset_ballots(self) -> None:
        contest_cvr_dl = self.get_cvr_dict(self._contest_rule_set_name)
        self._contest_cvr_ld = [{'ballot_marks': bm, 'weight': weight, 'weight_distrib': []}
                                for bm, weight in zip(contest_cvr_dl['ballot_marks'], contest_cvr_dl['weight'])]

    def _pre_check(self) -> None:
        """
        Any checks on the input data to make sure tabulation will be possible.
        """

        # check for all blank ballots, undervote or blank before exhaust
        ballot_sets = [b['ballot_marks'].unique_marks for b in self._contest_cvr_ld]
        if not set.union(*ballot_sets):
            raise RuntimeError(f"(tabulation={self._tab_num}) all effectively blank ballots")

    def _new_tabulation(self) -> None:
        """
        Add a new set of results for tabulation
        """
        self._tab_num += 1
        new_outcomes = {cand: {'name': cand, 'round_eliminated': None, 'round_elected': None}
                        for cand in self._contest_candidates.unique_candidates}
        self._tabulations.append(
            {
                'rounds': [],
                'summary_transfers': [],
                'by_candidate_transfers': [],
                'candidate_outcomes': new_outcomes,
                'final_weight_distrib': [],
                'final_ranks': [],
                'initial_ranks': [],
                'ballot_round_allocation': [],
                'ballot_round_weight': [],
                'win_threshold': None
            }
        )

    def _tabulate(self) -> None:
        """
        Run the rounds of rcv contest.
        """

        # use to mark first elimination round that occurs
        first_elimination_round = None

        # remove inactive candidates
        self._clean_ballots()

        # checks to make tabulation can proceed
        self._pre_check()

        # store initial values
        initial_ranks = [b['ballot_marks'].marks for b in self._contest_cvr_ld]
        self._tabulations[self._tab_num-1]['initial_ranks'] = initial_ranks

        not_complete = self._contest_not_complete()
        while not_complete:
            self._round_num += 1

            #############################################
            # CLEAR LAST ROUND VALUES
            self._round_winners = []
            self._round_loser = None

            #############################################
            # COUNT ROUND RESULTS
            self._tally_active_ballots()

            #############################################
            # CHECK FOR ROUND WINNERS
            self._set_round_winners()

            # on the first elimination round, mark any candidates with zero votes for elimination
            if first_elimination_round is None and not self._round_winners:
                round_dict = self.get_round_tally_dict(self._round_num, tabulation_num=self._tab_num)
                novote_losers = [cand for cand in self._contest_candidates.unique_candidates if round_dict[cand] == 0]

                for loser in novote_losers:
                    self._tabulations[self._tab_num-1]['candidate_outcomes'][loser]['round_eliminated'] = self._round_num

                self._inactive_candidates += novote_losers
                first_elimination_round = False

                self._clean_ballots()

            #############################################
            # IDENTIFY ROUND LOSER
            self._set_round_loser()

            #############################################
            # UPDATE inactive candidate list using round winner/loser
            self._update_candidates()

            # update complete flag
            not_complete = self._contest_not_complete()

            #############################################
            # UPDATE WEIGHTS
            # don't update if contest over
            if not_complete:
                self._update_weights()

            #############################################
            # CALC ROUND TRANSFER
            if not_complete:
                self._calc_round_transfer()
            else:
                self._tabulations[self._tab_num-1]['summary_transfers'].append(
                    {cand: util.NAN for cand in self._contest_candidates.unique_candidates.union({'exhaust'})})
                self._tabulations[self._tab_num-1]['by_candidate_transfers'].append({})

            #############################################
            # CLEAN ROUND BALLOTS
            # remove inactive candidates
            # don't clean if contest over
            if not_complete:
                self._clean_ballots()

        # record final ballot weight distributions
        final_weight_distrib = [b['weight_distrib'] + [(b['ballot_marks'].marks[0], b['weight'])]
                                if b['ballot_marks'].marks else b['weight_distrib'] + [('empty', b['weight'])]
                                for b in self._contest_cvr_ld]
        self._tabulations[self._tab_num-1]['final_weight_distrib'] = final_weight_distrib

        # set final ranks for each ballot
        final_ranks = [b['ballot_marks'].marks for b in self._contest_cvr_ld]
        self._tabulations[self._tab_num-1]['final_ranks'] = final_ranks

        self._tabulations[self._tab_num-1]['win_threshold'] = self._win_threshold()

    def _clean_ballots(self) -> None:
        """
        Remove any newly inactivated candidates from the ballot ranks.
        """
        for inactive_cand in self._inactive_candidates:
            if inactive_cand not in self._removed_candidates:
                self._contest_cvr_ld = [
                    {
                        'ballot_marks': BallotMarks.remove_mark(b['ballot_marks'], [inactive_cand]),
                        'weight': b['weight'],
                        'weight_distrib': b['weight_distrib']
                    }
                    for b in self._contest_cvr_ld]
                self._removed_candidates.append(inactive_cand)

    def _tally_active_ballots(self) -> None:

        # tally current and distributed weights
        ballot_alloc = []
        ballot_alloc_weight = []
        vote_alloc = collections.Counter({cand: 0 for cand in self._contest_candidates.unique_candidates})

        for b in self._contest_cvr_ld:

            candidate = "exhaust" if len(b['ballot_marks'].marks) == 0 else b['ballot_marks'].marks[0]
            ballot_alloc.append(candidate)
            ballot_alloc_weight.append(b['weight'])

            if candidate != "exhaust":
                vote_alloc[candidate] += b['weight']

            if b['weight_distrib']:
                for candidate, weight in b['weight_distrib']:
                    vote_alloc[candidate] += weight

        round_results = list(zip(*vote_alloc.most_common()))
        self._tabulations[self._tab_num-1]['rounds'].append(round_results)
        self._tabulations[self._tab_num-1]['ballot_round_allocation'].append(ballot_alloc)
        self._tabulations[self._tab_num-1]['ballot_round_weight'].append(ballot_alloc_weight)

    def _update_candidates(self) -> None:
        """
        Update candidate outcomes
        Assume winners are to become inactive, otherwise inactivate loser
        """

        # update winner outcomes
        for winner in self._round_winners:
            self._tabulations[self._tab_num-1]['candidate_outcomes'][winner]['round_elected'] = self._round_num
            self._inactive_candidates.append(winner)

        # if contest is not over
        if self._contest_not_complete():

            # if no winner, add loser
            if not self._round_winners:
                self._inactive_candidates.append(self._round_loser)
                self._tabulations[self._tab_num-1]['candidate_outcomes'][self._round_loser]['round_eliminated'] = self._round_num

        # if contest is over
        else:

            # set all remaining non-winners as eliminated
            remaining_candidates = [d['name'] for d in self._tabulations[self._tab_num-1]['candidate_outcomes'].values()
                                    if d['round_elected'] is None and d['round_eliminated'] is None]
            for cand in remaining_candidates:
                self._tabulations[self._tab_num-1]['candidate_outcomes'][cand]['round_eliminated'] = self._round_num
            self._inactive_candidates += remaining_candidates

    def _set_round_loser(self) -> None:
        """
        Find candidate from round with least votes.
        If more than one, choose randomly
        """

        # split round results into two tuples (index-matched)
        active_candidates, round_tallies = self.get_round_tally_tuple(self._round_num, self._tab_num,
                                                                      only_round_active_candidates=True, desc_sort=True)
        # find round loser
        # ignore zero vote candidates, they will be automtically eliminated with the first non-zero loser
        loser_count = min(i for i in round_tallies if i)

        # haven't implemented any special rules for tied losers. Print a warning if one is reached
        # if len([cand for cand, cand_tally
        #         in zip(active_candidates, round_tallies) if cand_tally == loser_count]) > 1:
        #     raise RuntimeWarning("reached a round with tied losers....")

        # in case of tied losers, choose one to eliminate (the last one in alpha order)
        round_losers = sorted([cand for cand, cand_tally
                               in zip(active_candidates, round_tallies)
                               if cand_tally == loser_count])
        self._round_loser = random.sample(round_losers, 1)[0]

    def get_round_tally_tuple(self,
                              round_num: int,
                              tabulation_num: int = 1,
                              only_round_active_candidates: bool = False,
                              desc_sort: bool = True) -> List[Tuple[str], Tuple[decimal.Decimal]]:
        """
        Return a dictionary containing keys as candidates and values as their vote counts in the round.
        """
        cands, tallies = self._tabulations[tabulation_num-1]['rounds'][round_num-1]

        # remove elected or eliminated candidates
        if only_round_active_candidates:

            outcomes = self._tabulations[tabulation_num-1]['candidate_outcomes']

            elected_filter = [(outcomes[cand]['round_elected'] is None or outcomes[cand]['round_elected'] >= round_num)
                              for cand in outcomes]
            eliminated_filter = [(outcomes[cand]['round_eliminated'] is None or outcomes[cand]['round_eliminated'] >= round_num)
                                 for cand in outcomes]

            active_candidates = [cand for cand, elect_filt, elim_filt in zip(outcomes, elected_filter, eliminated_filter)
                                 if elect_filt and elim_filt]
            tallies = [tally for idx, tally in enumerate(tallies) if cands[idx] in active_candidates]
            cands = [cand for cand in cands if cand in active_candidates]

        # sort
        rounds = list(zip(*[(cand, tally) for cand, tally in sorted(zip(cands, tallies), key=lambda x: -x[1])]))

           # rounds = [tuple(cands), tuple(tallies)]

        return rounds

    def get_round_tally_dict(self,
                             round_num: int,
                             tabulation_num: int = 1,
                             only_round_active_candidates: bool = False) -> Dict[str, decimal.Decimal]:
        """
        Return a dictionary containing keys as candidates and values as their vote counts in the round. Includes
        zero vote candidates and those winners remaining at threshold.
        """
        # convert to dict
        return {cand: count for cand, count in
                zip(*self.get_round_tally_tuple(round_num,
                                                tabulation_num,
                                                only_round_active_candidates=only_round_active_candidates))}

    def get_round_transfer_dict(self,
                                round_num: int,
                                candidate_netted: bool = True,
                                tabulation_num: int = 1) -> Dict[str, decimal.Decimal]:
        """
        Return a dictionary containing keys as candidates + 'exhaust' and values as their round net transfer
        """
        if candidate_netted:
            transfers = self._tabulations[tabulation_num-1]['summary_transfers']
        else:
            transfers = self._tabulations[tabulation_num-1]['by_candidate_transfers']
        return transfers[round_num-1]

    def get_candidate_outcomes(self, tabulation_num: int = 1) -> List[Dict]:
        """
        Return a list of dictionaries {keys: name, round_elected, round_eliminated}
        """
        candidate_outcomes = self._tabulations[tabulation_num-1]['candidate_outcomes']
        return list(candidate_outcomes.values())

    def get_final_weights(self, tabulation_num: int = 1) -> List[decimal.Decimal]:
        """
        Return a list of ballot weights after tabulation, index-matched with ballots
        """
        final_weights = self._tabulations[tabulation_num-1]['ballot_round_weight'][-1]
        return final_weights

    def get_initial_ranks(self, tabulation_num: int = 1) -> List[List]:
        """
        Return a list of ballot ranks prior to tabulation, but after an initial cleaning. Each set of ranks is a list.
        """
        initial_ranks = self._tabulations[tabulation_num-1]['initial_ranks']
        return initial_ranks

    def get_initial_weights(self, tabulation_num: int = 1) -> List[decimal.Decimal]:
        """
        Return a list of ballot weights prior to tabulation, but after an initial cleaning. Each set of ranks is a list.
        """
        initial_weights = self._tabulations[tabulation_num-1]['ballot_round_weight'][0]
        return initial_weights

    def get_final_ranks(self, tabulation_num: int = 1) -> List[List]:
        """
        Return a list of ballot ranks after tabulation. Each set of ranks is a list.
        """
        final_ranks = self._tabulations[tabulation_num-1]['final_ranks']
        return final_ranks

    def get_final_weight_distrib(self, tabulation_num: int = 1) -> List[List[Tuple[str, decimal.Decimal]]]:
        """
        Return a list of ballot weight distributions after tabulation. Each set of weight distributions
        is a ranking-weight tuple pair. Ballots that exhausted have the string 'empty' in the ranking position of
        the tuple.
        """
        final_weights = self._tabulations[tabulation_num-1]['final_weight_distrib']
        return final_weights

    def get_win_threshold(self, tabulation_num: int = 1) -> Optional[Union[int, float]]:
        return self._tabulations[tabulation_num-1]['win_threshold']

    def finalist_candidates(self, tabulation_num: int = 1) -> List[str]:
        """
        Return list of candidates with any ballot weight allotted to them by the end of tabulation.
        """
        final_weight_distrib = self.get_final_weight_distrib(tabulation_num=tabulation_num)
        final_weight_cands = list(set(t[0] for t in util.flatten_list(final_weight_distrib)).difference({'empty'}))
        return final_weight_cands

    def n_rounds(self, tabulation_num: int = 1) -> int:
        """
        Return the number of rounds, for a given tabulation.
        """
        rounds = self._tabulations[tabulation_num-1]['rounds']
        return len(rounds)

    def n_tabulations(self) -> int:
        return self._tab_num




from typing import (List)

import abc
import copy
import decimal

from rcv_cruncher.marks import BallotMarks
from rcv_cruncher.rcv.base import RCV


def get_rcv_dict():
    """
    Return dictionary of rcv classes, class_name: class_obj (constructor function)
    """
    return {
        'Until2': Until2,
        'STVWholeBallot': STVWholeBallot,
        'STVFractionalBallot': STVFractionalBallot,
        'Sequential': Sequential,
        'SingleWinner': SingleWinner,
        'BottomsUpThresh': BottomsUpThresh
    }


class SingleWinner(RCV):
    """
    Single winner rcv contest.
    - Winner is candidate to first achieve more than half the active round votes.
    - Votes are transferred from losers each round.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _contest_stats(self) -> List:
        """
        Every rcv variant must specify which stats list it uses.
        Available lists should be set as rcv base methods or reporting methods.
        """
        return self.single_winner_stats

    def _set_round_winners(self) -> None:
        """
        This function should set self._round_winners to the list of candidates that won the round

        single winner rules:
        - winner is candidate with > 50% of round vote
        """
        round_candidates, round_tallies = self.get_round_tally_tuple(self._round_num, self._tab_num,
                                                                     only_round_active_candidates=True, desc_sort=True)
        if round_tallies[0]*2 > sum(round_tallies):
            self._round_winners = [round_candidates[0]]

    def _calc_round_transfer(self) -> None:
        """
        This function should append a dictionary to self.transfers containing:
        candidate names as keys, plus one key for 'exhaust' and any other keys for transfer categories
        values as round transfer flows.

        rules:
        - transfer votes from round loser
        """
        candidates = self._contest_candidates.unique_candidates.union({'exhaust'})

        # calculate transfer
        by_candidate_transfer_dict = {
            from_cand: {to_cand: 0 for to_cand in candidates if to_cand != from_cand}
            for from_cand in candidates if from_cand != 'exhaust'
            }
        summary_transfer_dict = {cand: 0 for cand in candidates}

        for b in self._contest_cvr_ld:
            if len(b['ballot_marks'].marks) > 0 and b['ballot_marks'].marks[0] == self._round_loser:
                if len(b['ballot_marks'].marks) > 1:
                    transfer_to_candidate = b['ballot_marks'].marks[1]
                    summary_transfer_dict[transfer_to_candidate] += b['weight']
                    by_candidate_transfer_dict[self._round_loser][transfer_to_candidate] += b['weight']
                else:
                    summary_transfer_dict['exhaust'] += b['weight']
                    by_candidate_transfer_dict[self._round_loser]['exhaust'] += b['weight']

        summary_transfer_dict[self._round_loser] = sum(summary_transfer_dict.values()) * -1
        self._tabulations[self._tab_num-1]['summary_transfers'].append(summary_transfer_dict)

        # remove candidates with no transfer
        by_candidate_transfer_dict = {
            from_cand: {
                to_cand: transfer_count for to_cand, transfer_count in transfers.items()
                if transfer_count != 0
                }
            for from_cand, transfers in by_candidate_transfer_dict.items()
            if sum(transfers.values()) != 0
        }
        self._tabulations[self._tab_num-1]['by_candidate_transfers'].append(by_candidate_transfer_dict)

    def _contest_not_complete(self) -> bool:
        """
        This function should return True if another round should be evaluated and False
        is the contest should complete.

        single winner rules:
        - Stop contest once a winner is found.
        """
        candidate_outcomes = self._tabulations[self._tab_num-1]['candidate_outcomes']
        all_rounds_elected = [d['round_elected'] is not None for d in candidate_outcomes.values()]
        if any(all_rounds_elected):
            return False
        else:
            return True


class Sequential(SingleWinner):
    """
    Sequential RCV elections used to elect multiple winners. Winners are elected one-by-one in repeated single winner
    RCV elections, each time with the previous winners effectively treated as eliminated.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    # overwrite _run_contest to run multiple single winner elections
    def _run_contest(self) -> None:

        winners = []
        self._tab_num = 0
        self._tabulations = []

        # continue until the number of winners is reached OR until candidates run out
        while len(winners) != self._n_winners and len(self._contest_candidates.unique_candidates - set(winners)) != 0:

            self._new_tabulation()

            # reset inputs
            self._reset_ballots()

            # tabulation-level
            self._inactive_candidates = copy.copy(winners)  # mark previous iteration winners as inactive
            self._removed_candidates = []

            # round-level
            self._round_num = 0

            # calculate single winner rcv
            self._tabulate()

            # get winner and store results
            tabulation_outcomes = self._tabulations[self._tab_num-1]['candidate_outcomes']
            winners.append([d['name'] for d in tabulation_outcomes.values() if d['round_elected'] is not None][0])


class Until2(SingleWinner):
    """
    Run single winner contest all the way down to final two canidates.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _set_round_winners(self) -> None:
        """
        This function should set self._round_winners to the list of candidates that won the round

        single winner rules:
        - winner is candidate with more votes when there are only two candidates left
        """
        round_candidates, _ = self.get_round_tally_tuple(self._round_num, self._tab_num,
                                                         only_round_active_candidates=True, desc_sort=True)
        if len(round_candidates) == 2:
            self._round_winners = [round_candidates[0]]


class STV(RCV, abc.ABC):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _set_round_winners(self):
        """
        This function should set self._round_winners to the list of candidates that won the round

        rules:
        - candidate is winner when they receive more than threshold.
        - last #winners candidates are automatically elected.
        """

        # process of elimination?
        # If 2 candidates have been elected and two more are still needed BUT only 2 remain,
        # then they are automatically elected.
        candidate_outcomes = self._tabulations[self._tab_num-1]['candidate_outcomes']
        round_candidates, round_tally = self.get_round_tally_tuple(self._round_num, self._tab_num,
                                                                   only_round_active_candidates=True, desc_sort=True)
        n_remaining_candidates = len([d for d in candidate_outcomes.values()
                                      if d['round_elected'] is None and d['round_eliminated'] is None])
        n_elected_candidates = len([d for d in candidate_outcomes.values()
                                    if d['round_elected'] is not None])
        if n_remaining_candidates + n_elected_candidates == self._n_winners:
            self._round_winners = round_candidates
            return

        # any threshold winners?
        threshold = self._win_threshold()
        all_winners = [(cand, tally) for cand, tally in zip(round_candidates, round_tally)
                       if tally > threshold]
        all_winners = [cand for cand, _ in sorted(all_winners, key=lambda x: (-x[1], x[0]))]
        if all_winners and self._multi_winner_rounds:
            self._round_winners = all_winners
        elif all_winners:
            self._round_winners = [all_winners[0]]

    def _contest_not_complete(self) -> bool:
        """
        This function should return True if another round should be evaluated and False
        is the contest should complete.

        rules:
        - Stop once num_winners are elected.
        """
        candidate_outcomes = self._tabulations[self._tab_num-1]['candidate_outcomes']
        all_rounds_elected = [d['round_elected'] is not None for d in candidate_outcomes.values()]
        if sum(all_rounds_elected) == self._n_winners:
            return False
        else:
            return True

    def _win_threshold(self) -> decimal.Decimal:
        """
        rules:
        - # of votes in first round /(# to elect + 1)
        """
        if not self._tabulations[self._tab_num-1]['rounds']:
            print("Threshold depends on first round count. " +
                  "Don't call win_threshold() before first call to tally_active_ballots()")
            raise RuntimeError

        first_round_dict = self.get_round_tally_dict(1, tabulation_num=self._tab_num)
        first_round_active_votes = sum(first_round_dict.values())
        return int((first_round_active_votes / (self._n_winners + 1)) + 1)


class STVWholeBallot(STV):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        weights = set(b['weight'] for b in self._contest_cvr_ld)
        if len(weights) != 1:
            # see _removal_ballots
            raise RuntimeError('ballots with unequal weights will not work with current implementation')

    def _removal_ballots(self, candidate, default_as_true=True) -> List[bool]:

        if default_as_true:
            # marks all ballots to be transferred
            to_remove = [True] * len(self._contest_cvr_ld)
        else:
            to_remove = [False] * len(self._contest_cvr_ld)

        # all ballots are transferred for non-winners
        if candidate not in self._round_winners:
            return to_remove

        thresh = self._win_threshold()
        ballot_weight = self._contest_cvr_ld[0]['weight']  # assuming equal weighted ballots
        continuing_candidates = set(self._contest_candidates) - set(self._inactive_candidates)

        # total and surplus
        cand_total = self.get_round_tally_dict(self._round_num, self._tab_num)[candidate]
        surplus = cand_total - thresh

        # skip factor for determining transferred ballots
        surplus_factor = int(round(cand_total/surplus))

        # pull out ballots that counted towards this winning candidate
        cand_ballots = [(idx, b) for idx, b in enumerate(self._contest_cvr_ld)
                        if b['ballot_marks'].marks and b['ballot_mark'].marks[0] == candidate]

        # mark all ballot belonging to candidate as not-transfer
        for ballot in cand_ballots:
            to_remove[ballot[0]] = False

        seen_idxs = []
        used_idxs = []
        for offset in range(0, surplus_factor):

            # start_idx = offset
            start_idx = surplus_factor + offset - 1
            skip_factor = surplus_factor

            print((start_idx, len(cand_ballots), skip_factor))
            for idx in range(start_idx, len(cand_ballots), skip_factor):

                print(idx)

                if idx in seen_idxs or idx in used_idxs:
                    raise RuntimeError("should this happen?")

                # enough ballots have been marked for transfer
                if cand_total <= thresh:
                    break

                # if current ballot has a next choice that is a continuing candidate,
                ranks = cand_ballots[idx][1]['ranks']
                if set(ranks).intersection(continuing_candidates):
                    # mark it for transfer
                    to_remove[cand_ballots[idx][0]] = True
                    # decrement candidate total
                    cand_total -= ballot_weight
                    used_idxs.append(idx)

                seen_idxs.append(idx)

        if cand_total > thresh:

            unseen_idx = [idx for idx, _ in enumerate(cand_ballots) if idx not in seen_idxs]
            non_transfer_continuing = [bool(set(b[1]['ballot_marks'].marks).intersection(continuing_candidates))
                                       for idx, b in enumerate(cand_ballots) if idx not in used_idxs]
            non_transfer_ballot_idx = [b[0] for idx, b in enumerate(cand_ballots) if idx not in used_idxs]

            if any(flag for idx, flag in enumerate(non_transfer_continuing) if idx not in unseen_idx):
                print("some un-chosen ballots have continuing candidates")
                raise RuntimeError
            else:
                for idx in non_transfer_ballot_idx:
                    if cand_total <= thresh:
                        break
                    to_remove[idx] = True
                    cand_total -= ballot_weight

        if cand_total > thresh:
            raise RuntimeError

        return to_remove

    def _calc_round_transfer(self) -> None:
        """
        This function should append a dictionary to self.transfers containing:
        candidate names as keys, plus one key for 'exhaust' and any other keys for transfer categories
        values as round transfer flows.

        rules:
        - transfer votes from round loser or winner
        """

        candidates = self._contest_candidates.unique_candidates.union({'exhaust'})

        # calculate transfer
        by_candidate_transfer_dict = {
            from_cand: {to_cand: 0 for to_cand in candidates if to_cand != from_cand}
            for from_cand in candidates if from_cand != 'exhaust'
            }
        summary_transfer_dict = {cand: 0 for cand in candidates}

        if self._round_winners:

            # collect flags indicating
            all_removal_ballots = []
            for winner in self._round_winners:
                all_removal_ballots.append(self._removal_ballots(winner, default_as_true=False))

            combined_removal_ballots = [False] * len(self._contest_cvr_ld)
            for idx in range(0, len(self._contest_cvr_ld)):
                if any(flag_list[idx] for flag_list in all_removal_ballots):
                    combined_removal_ballots[idx] = True

            for b, is_transfer in zip(self._contest_cvr_ld, combined_removal_ballots):

                if is_transfer:

                    eliminated_candidate = b['ballot_marks'].marks[0]

                    if len(b['ballot_marks'].marks) == 0:
                        print('this shouldnt happen.')
                        raise RuntimeError

                    remaining_candidates = [cand for cand in b['ballot_marks'].marks
                                            if cand not in self._round_winners]
                    if len(remaining_candidates) == 0:
                        # exhausted
                        summary_transfer_dict['exhaust'] += b['weight']
                        by_candidate_transfer_dict[eliminated_candidate]['exhaust'] += b['weight']
                    elif len(remaining_candidates) > 0:
                        # transfer to another candidate
                        transfer_to = remaining_candidates[0]
                        summary_transfer_dict[transfer_to] += b['weight']
                        by_candidate_transfer_dict[eliminated_candidate][transfer_to] += b['weight']

                    # include outflow
                    summary_transfer_dict[eliminated_candidate] += b['weight'] * -1

        else:
            transfer_candidates = [self._round_loser]

            # calculate transfer
            for b in self._contest_cvr_ld:
                if len(b['ballot_marks'].marks) > 0 and b['ballot_marks'].marks[0] in transfer_candidates:

                    eliminated_candidate = b['ballot_marks'].marks[0]

                    if len(b['ballot_marks'].marks) > 1:
                        transfer_to = b['ballot_marks'].marks[1]
                        summary_transfer_dict[transfer_to] += b['weight']
                        by_candidate_transfer_dict[eliminated_candidate][transfer_to] += b['weight']
                    else:
                        summary_transfer_dict['exhaust'] += b['weight']
                        by_candidate_transfer_dict[eliminated_candidate]['exhaust'] += b['weight']

                    # mark transfer outflow
                    summary_transfer_dict[eliminated_candidate] += b['weight'] * -1

        self._tabulations[self._tab_num-1]['summary_transfers'].append(summary_transfer_dict)

        # remove candidates with no transfer
        by_candidate_transfer_dict = {
            from_cand: {
                to_cand: transfer_count for to_cand, transfer_count in transfers.items()
                if transfer_count != 0
                }
            for from_cand, transfers in by_candidate_transfer_dict.items()
            if sum(transfers.values()) != 0
        }
        self._tabulations[self._tab_num-1]['by_candidate_transfers'].append(by_candidate_transfer_dict)


    def _clean_ballots(self) -> None:
        """
        Remove any newly inactivated candidates from the ballot ranks. But only remove previous round winners
        from specified ballots
        """
        winners = []
        for inactive_cand in self._inactive_candidates:
            if inactive_cand not in self._removed_candidates:
                if inactive_cand in self._round_winners:
                    winners.append(inactive_cand)
                    self._removed_candidates.append(inactive_cand)
                else:
                    self._contest_cvr_ld = [
                        {
                            'ballot_marks': BallotMarks.remove_mark(b['ballot_marks'], [inactive_cand]),
                            'weight': b['weight'],
                            'weight_distrib': b['weight_distrib']
                        }
                        for b in self._contest_cvr_ld]
                    self._removed_candidates.append(inactive_cand)

        # remove all other candidates before winners. This mostly matters in the first round when all
        # zero-vote candidates are removed. That is the only time a loser and a winner might both be inactivated in the
        # same round. It may also happen in the last round, but transfer calculations at that point have no impact.
        remove_bool_lists = []
        for winner in winners:
            remove_bool_lists.append(self._removal_ballots(winner, default_as_true=True))

        for winner, to_remove in zip(winners, remove_bool_lists):
            self._contest_cvr_ld = [
                {
                    'ballot_marks': BallotMarks.remove_mark(b['ballot_marks'], [inactive_cand]),
                    'weight': b['weight'],
                    'weight_distrib': b['weight_distrib']
                }
                if is_remove else b
                for b, is_remove in zip(self._contest_cvr_ld, to_remove)]


class STVFractionalBallot(STV):
    """
    Multi-winner elections with fractional ballot transfer.
    - Win threshold is set as the (# first round votes)/(# of seats + 1).
    - Any winner is eliminated and has their surplus redistributed. Percent of each
    ballot redistributed is equal to (# of votes winner has -- minus the threshold)/(# of votes winner has).
    - If no winners in round, candidate with least votes in a round is eliminated and has votes transferred.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _update_weights(self) -> None:
        """
        If surplus needs to be transferred, change weights on winner ballots to reflect remaining
        active ballot proportion.

        rules:
        - reduce weights of ballots ranking the winner by the amount
        """

        round_dict = self.get_round_tally_dict(self._round_num, tabulation_num=self._tab_num)
        threshold = self._win_threshold()

        for winner in self._round_winners:

            # fractional surplus to transfer from each winner ballot
            surplus_percent = (round_dict[winner] - threshold) / round_dict[winner]

            # if surplus to transfer is non-zero
            if surplus_percent:

                # which ballots had the winner on top
                # and need to be fractionally split
                new = []
                for b in self._contest_cvr_ld:
                    if b['ballot_marks'].marks and b['ballot_marks'].marks[0] == winner:

                        # record ballot weight allotted to winner
                        winner_weight = b['weight'] * (1 - surplus_percent)
                        new_weight_distrib = b['weight_distrib'] + [(winner, winner_weight)]

                        # remaining weight
                        remaining_weight = b['weight'] * surplus_percent

                        # adjust ballot's current weight
                        new.append(
                            {
                                'ballot_marks': b['ballot_marks'],
                                'weight': remaining_weight,
                                'weight_distrib': new_weight_distrib
                            })
                    else:
                        new.append(b)
                self._contest_cvr_ld = new

    #
    def _calc_round_transfer(self) -> None:
        """
        This function should append a dictionary to self.transfers containing:
        candidate names as keys, plus one key for 'exhaust' and any other keys for transfer categories
        values as round transfer flows.

        rules:
        - transfer votes from round loser or winner
        """
        if self._round_winners:
            transfer_candidates = self._round_winners
        else:
            transfer_candidates = [self._round_loser]

        candidates = self._contest_candidates.unique_candidates.union({'exhaust'})

        # calculate transfer
        by_candidate_transfer_dict = {
            from_cand: {to_cand: 0 for to_cand in candidates if to_cand != from_cand}
            for from_cand in candidates if from_cand != 'exhaust'
            }
        summary_transfer_dict = {cand: 0 for cand in candidates}

        for b in self._contest_cvr_ld:
            if len(b['ballot_marks'].marks) > 0 and b['ballot_marks'].marks[0] in transfer_candidates:

                eliminated_candidate = b['ballot_marks'].marks[0]

                remaining_candidates = [cand for cand in b['ballot_marks'].marks
                                        if cand not in transfer_candidates]
                if remaining_candidates:
                    transfer_to = remaining_candidates[0]
                    summary_transfer_dict[transfer_to] += b['weight']
                    by_candidate_transfer_dict[eliminated_candidate][transfer_to] += b['weight']
                else:
                    summary_transfer_dict['exhaust'] += b['weight']
                    by_candidate_transfer_dict[eliminated_candidate]['exhaust'] += b['weight']

                # mark transfer outflow
                summary_transfer_dict[eliminated_candidate] += b['weight'] * -1

        self._tabulations[self._tab_num-1]['summary_transfers'].append(summary_transfer_dict)

        # remove candidates with no transfer
        by_candidate_transfer_dict = {
            from_cand: {
                to_cand: transfer_count for to_cand, transfer_count in transfers.items()
                if transfer_count != 0
                }
            for from_cand, transfers in by_candidate_transfer_dict.items()
            if sum(transfers.values()) != 0
        }
        self._tabulations[self._tab_num-1]['by_candidate_transfers'].append(by_candidate_transfer_dict)


class BottomsUpThresh(RCV):
    """
    Multi winner contest. When all candidates in a round have more than X% of the round votes, they are all winners.
    - In a no winner round, the candidate with least votes is eliminated and ballots transferred.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self._bottoms_up_threshold is None:
            raise RuntimeError('BottomsUpThresh contest cannot have "bottoms_up_thresh" value "None"')

    def _set_round_winners(self) -> None:
        """
        This function should set self._round_winners to the list of candidates that won the round
        """
        round_candidates, round_tallies = self.get_round_tally_tuple(self._round_num, self._tab_num,
                                                                     only_round_active_candidates=True)
        threshold = sum(round_tallies) * self._bottoms_up_threshold
        if all(i > threshold for i in round_tallies):
            self._round_winners = list(round_candidates)

    def _calc_round_transfer(self) -> None:
        """
        This function should append a dictionary to self.transfers containing:
        candidate names as keys, plus one key for 'exhaust' and any other keys for transfer categories
        values as round transfer flows.

        rules:
        - transfer votes from round loser
        """

        candidates = self._contest_candidates.unique_candidates.union({'exhaust'})

        # calculate transfer
        by_candidate_transfer_dict = {
            from_cand: {to_cand: 0 for to_cand in candidates if to_cand != from_cand}
            for from_cand in candidates if from_cand != 'exhaust'
            }
        transfer_dict = {cand: 0 for cand in self._candidate_set.union({'exhaust'})}

        for b in self._contest_cvr_ld:
            if len(b['ballot_marks'].marks) > 0 and b['ballot_marks'].marks[0] == self._round_loser:
                if len(b['ballot_marks'].marks) > 1:
                    transfer_to = b['ballot_marks'].marks[1]
                    transfer_dict[transfer_to] += b['weight']
                    by_candidate_transfer_dict[self._round_loser][transfer_to] += b['weight']
                else:
                    transfer_dict['exhaust'] += b['weight']
                    by_candidate_transfer_dict[self._round_loser]['exhaust'] += b['weight']

        transfer_dict[self._round_loser] = sum(transfer_dict.values()) * -1
        self._tabulations[self._tab_num-1]['summary_transfers'].append(transfer_dict)

        # remove candidates with no transfer
        by_candidate_transfer_dict = {
            from_cand: {
                to_cand: transfer_count for to_cand, transfer_count in transfers.items()
                if transfer_count != 0
                }
            for from_cand, transfers in by_candidate_transfer_dict.items()
            if sum(transfers.values()) != 0
        }
        self._tabulations[self._tab_num-1]['by_candidate_transfers'].append(by_candidate_transfer_dict)

    def _contest_not_complete(self) -> None:
        """
        This function should return True if another round should be evaluated and False
        is the contest should complete.

        single winner rules:
        - Contest is over when all winner are found, they are all 'elected' at once.
        """
        candidate_outcomes = self._tabulations[self._tab_num-1]['candidate_outcomes']
        all_rounds_elected = [d['round_elected'] is not None for d in candidate_outcomes.values()]
        if any(all_rounds_elected):
            return False
        else:
            return True




import collections
import weightedstats

import pandas as pd

import rcv_cruncher.util as util

from rcv_cruncher.marks import BallotMarks


class CastVoteRecord_stats:

    def stats(self,
              keep_decimal_type: bool = False,
              add_split_stats: bool = False,
              add_id_info: bool = True) -> pd.DataFrame:

        cvr_stats = self._summary_cvr_stat_table.copy()

        if add_id_info:
            for col in self._id_df.columns[::-1]:
                cvr_stats.insert(0, col, self._id_df[col].item())

        if add_split_stats:

            self._make_split_filter_dict()
            self._compute_summary_cvr_split_stat_table()

            if self._summary_cvr_split_stat_table is not None:

                cvr_split_stats = self._summary_cvr_split_stat_table.copy()

                cvr_split_stats = cvr_split_stats.assign(**{col: cvr_stats.at[0, col] for col in cvr_stats.columns})
                cvr_split_stats = cvr_split_stats[
                    cvr_split_stats.columns.tolist()[-1 * len(cvr_stats.columns):] +
                    cvr_split_stats.columns.tolist()[:-1 * len(cvr_stats.columns)]
                ]

                cvr_stats = cvr_split_stats

        if not keep_decimal_type:
            cvr_stats = cvr_stats.applymap(util.decimal2float)

        return cvr_stats

    def _compute_cvr_stat_table(self) -> None:

        cvr = self.get_cvr_dict()
        candidates = self.get_candidates()

        df = pd.DataFrame()
        df['weight'] = cvr['weight']

        df['valid_ranks_used'] = [len(b.unique_candidates) for b in cvr['ballot_marks']]
        df['ranks_used_times_weight'] = df['valid_ranks_used'] * df['weight']

        df['used_last_rank'] = [True if b.marks[-1] != BallotMarks.SKIPPED else False for b in cvr['ballot_marks']]

        df['undervote'] = [b.unique_marks == {BallotMarks.SKIPPED} for b in cvr['ballot_marks']]
        df['ranked_single'] = df['valid_ranks_used'] == 1
        df['ranked_multiple'] = df['valid_ranks_used'] > 1
        df['ranked_3_or_more'] = df['valid_ranks_used'] > 2

        ballot_marks_no_skipped = [BallotMarks.remove_mark(b, [BallotMarks.SKIPPED]) for b in cvr['ballot_marks']]
        first_round = [b.marks[0] if b.marks else 'NA' for b in ballot_marks_no_skipped]
        df['first_round'] = pd.Series(first_round, dtype='category')

        df['first_round_overvote'] = df['first_round'].eq(BallotMarks.OVERVOTE)

        df['contains_overvote'] = [BallotMarks.OVERVOTE in b.unique_marks for b in cvr['ballot_marks']]

        # contains_skipped
        # {SKIPVOTE} & {x} - {y}
        # this checks that x == SKIPVOTE and that y then != SKIPVOTE
        # (the y check is important to know whether or not the ballot contains marks
        # following the skipped rank)
        df['contains_skip'] = [any({BallotMarks.SKIPPED} & {x} - {y} for x, y in zip(b.marks, b.marks[1:]))
                               for b in cvr['ballot_marks']]

        # contains_duplicate
        # remove overvotes and undervotes
        dup_check = [BallotMarks.remove_mark(b, [BallotMarks.SKIPPED, BallotMarks.OVERVOTE]) for b in cvr['ballot_marks']]
        # count all ranks for candidates
        counters = [collections.Counter(b.marks) for b in dup_check]
        # check if any candidates were ranked more than once
        df['contains_duplicate'] = [max(counter.values()) > 1 if counter else False for counter in counters]

        irregular_condtions = ['contains_overvote', 'contains_skip', 'contains_duplicate']
        df['irregular'] = df[irregular_condtions].any(axis='columns')

        # fully_ranked no overvotes
        candidates_combined_writeins = BallotMarks.combine_writein_marks(candidates)
        candidates_excluded_writeins = BallotMarks.remove_mark(candidates_combined_writeins, [BallotMarks.WRITEIN])
        candidate_set = candidates_excluded_writeins.unique_candidates

        ballot_marks_cleaned = [BallotMarks.remove_mark(b, [BallotMarks.OVERVOTE, BallotMarks.SKIPPED])
                                for b in cvr['ballot_marks']]
        ballot_marks_cleaned = [BallotMarks.remove_duplicate_candidate_marks(b) for b in ballot_marks_cleaned]

        fully_ranked = [(set(b.marks) & candidate_set) == candidate_set or
                        # voters ranked every possible candidate
                        len(a.marks) == len(b.marks)
                        # or did not, had no skipped ranks, overvotes, or duplicates
                        for a, b in zip(cvr['ballot_marks'], ballot_marks_cleaned)]
        df['fully_ranked_excl_overvotes'] = fully_ranked

        # fully_ranked with overvotes
        candidates_combined_writeins = BallotMarks.combine_writein_marks(candidates)
        candidates_excluded_writeins = BallotMarks.remove_mark(candidates_combined_writeins, [BallotMarks.WRITEIN])
        candidate_set = candidates_excluded_writeins.unique_candidates

        ballot_marks_cleaned = [BallotMarks.remove_mark(b, [BallotMarks.SKIPPED])
                                for b in cvr['ballot_marks']]
        ballot_marks_cleaned = [BallotMarks.remove_duplicate_candidate_marks(b) for b in ballot_marks_cleaned]

        fully_ranked = [(set(b.marks) & candidate_set) == candidate_set or
                        # voters ranked every possible candidate
                        len(a.marks) == len(b.marks)
                        # or did not, had no skipped ranks, overvotes, or duplicates
                        for a, b in zip(cvr['ballot_marks'], ballot_marks_cleaned)]
        df['fully_ranked_incl_overvotes'] = fully_ranked

        self._cvr_stat_table = df

    def _compute_summary_cvr_stat_table(self) -> None:

        cvr = self.get_cvr_dict()
        candidates = self.get_candidates()

        s = pd.Series(dtype=object)

        candidates_no_writeins = BallotMarks.remove_mark(BallotMarks.combine_writein_marks(candidates), [BallotMarks.WRITEIN])
        s['n_candidates'] = len(candidates_no_writeins.marks)

        s['rank_limit'] = len(cvr['ballot_marks'][0].marks)
        s['restrictive_rank_limit'] = True if s['rank_limit'] < (s['n_candidates'] - 1) else False

        # first_round_overvote
        # The number of ballots with an overvote before any valid ranking. (weighted)

        # Note that this is not the same as "exhausted by overvote". This is because
        # some jurisdictions (Maine) discard any ballot beginning with two
        # skipped rankings, and call this ballot as exhausted by skipped rankings, even if the
        # skipped rankings are followed by an overvote.

        # Other jursidictions (Minneapolis) simply skip over overvotes in a ballot.
        s['first_round_overvote'] = self._cvr_stat_table.loc[self._cvr_stat_table['first_round_overvote'], 'weight'].sum()

        # The number of voters that validly used only a single ranking. (weighted)
        s['ranked_single'] = self._cvr_stat_table.loc[self._cvr_stat_table['ranked_single'], 'weight'].sum()

        # The number of voters that validly used 3 or more rankings. (weighted)
        s['ranked_3_or_more'] = self._cvr_stat_table.loc[self._cvr_stat_table['ranked_3_or_more'], 'weight'].sum()

        # The number of voters that validly use more than one ranking. (weighted)
        s['ranked_multiple'] = self._cvr_stat_table.loc[self._cvr_stat_table['ranked_multiple'], 'weight'].sum()

        # The number of voters that have validly used all available rankings on the
        # ballot, or that have validly ranked all non-write-in candidates. (weighted)
        s['total_fully_ranked'] = self._cvr_stat_table.loc[self._cvr_stat_table['fully_ranked_excl_overvotes'], 'weight'].sum()

        # The number of ballots that rank the same candidate more than once. (weighted)
        s['includes_duplicate_ranking'] = self._cvr_stat_table.loc[self._cvr_stat_table['contains_duplicate'], 'weight'].sum()

        # The number of ballots that have an skipped ranking followed by any other marked ranking. (weighted)
        s['includes_skipped_ranking'] = self._cvr_stat_table.loc[self._cvr_stat_table['contains_skip'], 'weight'].sum()

        # This includes ballots with no marks. (weighted)
        s['total_ballots'] = self._cvr_stat_table['weight'].sum()

        # Number of ballots that either had a multiple ranking, overvote,
        # or a skipped ranking (only those followed by a mark). This includes ballots even where the irregularity was not
        # the cause of exhaustion. (weighted)
        s['total_irregular'] = self._cvr_stat_table.loc[self._cvr_stat_table['irregular'], 'weight'].sum()

        # Number of ballots with at least one overvote. Not necessarily cause of exhaustion. (weighted)
        s['includes_overvote_ranking'] = self._cvr_stat_table.loc[self._cvr_stat_table['contains_overvote'], 'weight'].sum()

        # Ballots completely made up of skipped rankings (no marks). (weighted)
        s['total_undervote'] = self._cvr_stat_table.loc[self._cvr_stat_table['undervote'], 'weight'].sum()

        # Mean number of validly used rankings across all non-undervote ballots. (weighted)
        weighted_sum = self._cvr_stat_table.loc[~self._cvr_stat_table['undervote'], 'ranks_used_times_weight'].sum()
        s['mean_rankings_used'] = weighted_sum / self._cvr_stat_table.loc[~self._cvr_stat_table['undervote'], 'weight'].sum()

        # Median number of validly used rankings across all non-undervote ballots. (weighted)
        # s['median_rankings_used'] = self._cvr_stat_table.loc[~self._cvr_stat_table['undervote'], 'ranks_used_times_weight'].median()

        ranks_used = self._cvr_stat_table.loc[~self._cvr_stat_table['undervote'], 'valid_ranks_used'].tolist()
        weights = self._cvr_stat_table.loc[~self._cvr_stat_table['undervote'], 'weight'].tolist()
        weights_float = [float(i) for i in weights]
        s['median_rankings_used'] = weightedstats.weighted_median(ranks_used, weights=weights_float)

        self._summary_cvr_stat_table = s.to_frame().transpose()

    def _make_split_filter_dict(self) -> None:

        if self.split_fields:

            cvr = self.get_cvr_dict()
            cvr_fields = list(cvr.keys())
            field_name_lower_dict = {k.lower(): k for k in cvr_fields}

            for field in self.split_fields:
                if field.lower() in field_name_lower_dict:
                    cvr_field_name = field_name_lower_dict[field.lower()]
                    cvr_field = cvr[cvr_field_name]
                    field_filter_dict = {unique_val: [unique_val == i for i in cvr_field] for unique_val in set(cvr_field)}
                    self._split_filter_dict.update({cvr_field_name: field_filter_dict})

    def _clean_string(self, x: str) -> str:
        return str(x).replace(":", "_").replace("/", "_").replace("\\", "_").replace(" ", "_").replace("-", "_")

    def _compute_cvr_split_stats(self, split_filter) -> pd.DataFrame:

        filtered_stat_table = self._cvr_stat_table.loc[split_filter, :]

        first_round_overvote = filtered_stat_table.loc[filtered_stat_table['first_round_overvote'], 'weight'].sum()
        ranked_single = filtered_stat_table.loc[filtered_stat_table['ranked_single'], 'weight'].sum()
        ranked_multiple = filtered_stat_table.loc[filtered_stat_table['ranked_multiple'], 'weight'].sum()
        ranked_3_or_more = filtered_stat_table.loc[filtered_stat_table['ranked_3_or_more'], 'weight'].sum()
        total_fully_ranked = filtered_stat_table.loc[filtered_stat_table['fully_ranked_excl_overvotes'], 'weight'].sum()
        includes_duplicate_ranking = filtered_stat_table.loc[filtered_stat_table['contains_duplicate'], 'weight'].sum()
        includes_skipped_ranking = filtered_stat_table.loc[filtered_stat_table['contains_skip'], 'weight'].sum()
        total_irregular = filtered_stat_table.loc[filtered_stat_table['irregular'], 'weight'].sum()
        total_ballots = filtered_stat_table['weight'].sum()
        includes_overvote_ranking = filtered_stat_table.loc[filtered_stat_table['contains_overvote'], 'weight'].sum()
        total_undervote = filtered_stat_table.loc[filtered_stat_table['undervote'], 'weight'].sum()

        weighted_sum = filtered_stat_table.loc[~filtered_stat_table['undervote'], 'ranks_used_times_weight'].sum()
        mean_rankings_used = weighted_sum / filtered_stat_table.loc[~filtered_stat_table['undervote'], 'weight'].sum()

        ranks_used = filtered_stat_table.loc[~filtered_stat_table['undervote'], 'valid_ranks_used'].tolist()
        weights = filtered_stat_table.loc[~filtered_stat_table['undervote'], 'weight'].tolist()
        weights_float = [float(i) for i in weights]
        median_rankings_used = weightedstats.weighted_median(ranks_used, weights=weights_float)

        filtered_summary_stat_table = pd.DataFrame({
            'split_first_round_overvote': [first_round_overvote],
            'split_ranked_single': [ranked_single],
            'split_ranked_multiple': [ranked_multiple],
            'split_ranked_3_or_more': [ranked_3_or_more],
            'split_mean_rankings_used': [mean_rankings_used],
            'split_median_rankings_used': [median_rankings_used],
            'split_total_fully_ranked': [total_fully_ranked],
            'split_includes_duplicate_ranking': [includes_duplicate_ranking],
            'split_includes_skipped_ranking': [includes_skipped_ranking],
            'split_total_irregular': [total_irregular],
            'split_total_ballots': [total_ballots],
            'split_includes_overvote_ranking': [includes_overvote_ranking],
            'split_total_undervote': [total_undervote]
            })
        return filtered_summary_stat_table

    def _compute_summary_cvr_split_stat_table(self) -> None:

        if not self._split_filter_dict:
            return

        split_df_list = []

        for field in self._split_filter_dict:
            field_clean = self._clean_string(field)

            for unique_val in self._split_filter_dict[field]:
                val_clean = self._clean_string(unique_val)
                split_id = field_clean + "-" + val_clean

                split_id_df = pd.DataFrame({
                    'split_field': [field],
                    'split_value': [unique_val],
                    'split_id': [split_id]
                    })
                split_stat_df = self._compute_cvr_split_stats(self._split_filter_dict[field][unique_val])
                split_df_list.append(pd.concat([split_id_df, split_stat_df], axis='columns'))

        self._summary_cvr_split_stat_table = pd.concat(split_df_list, axis=0, sort=False)



from __future__ import annotations
from typing import (Callable, Dict, Optional, List, Type, Union, Tuple)

import decimal
import collections
import re
import pathlib

import pandas as pd

from rcv_cruncher.marks import BallotMarks
from rcv_cruncher.cvr.tables import CastVoteRecord_tables
from rcv_cruncher.cvr.stats import CastVoteRecord_stats

decimal.getcontext().prec = 30


class CastVoteRecord(CastVoteRecord_stats, CastVoteRecord_tables):

    @staticmethod
    def get_stats(cvr: Type[CastVoteRecord],
                  keep_decimal_type: bool = False,
                  add_split_stats: bool = False,
                  add_id_info: bool = True) -> pd.DataFrame:
        return cvr.stats(keep_decimal_type=keep_decimal_type,
                         add_split_stats=add_split_stats,
                         add_id_info=add_id_info)

    @staticmethod
    def write_cvr_table(cvr: Type[CastVoteRecord],
                        table_format: str = "rank",
                        save_dir: Union[str, pathlib.Path] = None) -> None:
        uid = cvr.stats()[0]['unique_id'].item()
        save_path = pathlib.Path(save_dir) / f'{uid}.csv'
        cvr.cvr_table(table_format=table_format).to_csv(save_path, index=False)

    @staticmethod
    def get_cumulative_ranking_tables(cvr: Type[CastVoteRecord]) -> Tuple[pd.DataFrame]:
        return cvr.cumulative_ranking_tables()

    @staticmethod
    def write_cumulative_ranking_tables(cvr: Type[CastVoteRecord],
                                        save_dir: Union[str, pathlib.Path] = None) -> None:
        count_df, percent_df = cvr.cumulative_ranking_tables()
        uid = cvr.stats()[0]['unique_id'].item()

        save_path = pathlib.Path(save_dir) / 'cumulative_ranking'
        save_path.mkdir(exist_ok=True)

        count_df.to_csv(save_path / f'{uid}_count.csv')
        percent_df.to_csv(save_path / f'{uid}_percent.csv')

    @staticmethod
    def get_first_second_tables(cvr: Type[CastVoteRecord]) -> Tuple[pd.DataFrame]:
        return cvr.first_second_tables()

    @staticmethod
    def write_first_second_tables(cvr: Type[CastVoteRecord],
                                  save_dir: Union[str, pathlib.Path] = None) -> None:
        count_df, percent_df, percent_no_exhaust_df = cvr.first_second_tables()
        uid = cvr.stats()[0]['unique_id'].item()

        save_path = pathlib.Path(save_dir) / 'first_second_choices'
        save_path.mkdir(exist_ok=True)

        count_df.to_csv(save_path / f'{uid}_count.csv')
        percent_df.to_csv(save_path / f'{uid}_percent.csv')
        percent_no_exhaust_df.to_csv(save_path / f'{uid}_percent_no_exhaust.csv')

    @staticmethod
    def get_rank_usage_table(cvr: Type[CastVoteRecord]) -> Tuple[pd.DataFrame]:
        return cvr.rank_usage_table()

    @staticmethod
    def write_rank_usage_table(cvr: Type[CastVoteRecord],
                               save_dir: Union[str, pathlib.Path] = None) -> None:
        df = cvr.rank_usage_table()
        uid = cvr.stats()[0]['unique_id'].item()

        save_path = pathlib.Path(save_dir) / 'rank_usage'
        save_path.mkdir(exist_ok=True)

        df.to_csv(save_path / f'{uid}.csv')

    @staticmethod
    def get_crossover_table(cvr: Type[CastVoteRecord]) -> pd.DataFrame:
        return cvr.crossover_table()

    @staticmethod
    def write_crossover_table(cvr: Type[CastVoteRecord],
                              save_dir: Union[str, pathlib.Path] = None) -> None:
        count_df, percent_df = cvr.crossover_tables()
        uid = cvr.stats()[0]['unique_id'].item()

        save_path = pathlib.Path(save_dir) / 'opponent_crossover'
        save_path.mkdir(exist_ok=True)

        count_df.to_csv(save_path / f'{uid}_count.csv')
        percent_df.to_csv(save_path / f'{uid}_percent.csv')

    @staticmethod
    def get_condorcet_tables(cvr: Type[CastVoteRecord]) -> Tuple[pd.DataFrame]:
        return cvr.condorcet_tables()

    @staticmethod
    def write_condorcet_tables(cvr: Type[CastVoteRecord],
                               save_dir: Union[str, pathlib.Path] = None) -> None:
        count_df, percent_df, condorcet_winner = cvr.condorcet_tables()
        uid = cvr.stats()[0]['unique_id'].item()

        save_path = pathlib.Path(save_dir) / 'condorcet'
        save_path.mkdir(exist_ok=True)

        count_df.to_csv(save_path / f'{uid}_count.csv', index_label=f'condorcet winner: {condorcet_winner}')
        percent_df.to_csv(save_path / f'{uid}_percent.csv', index_label=f'condorcet winner: {condorcet_winner}')

    def __init__(self,
                 jurisdiction: str = "",
                 state: str = "",
                 year: str = "",
                 date: str = "",
                 office: str = "",
                 notes: str = "",
                 parser_func: Optional[Callable] = None,
                 parser_args: Optional[Dict] = None,
                 parsed_cvr: Optional[Dict] = None,
                 split_fields: Optional[List] = None) -> None:

        # ID INFO
        self.jurisdiction = jurisdiction
        self.state = state
        self.date = date
        self.year = year
        self.office = office
        self.notes = notes
        self.split_fields = split_fields
        self.unique_id = self._unique_id()

        self._id_df = pd.DataFrame({
            'jurisdiction': [self.jurisdiction],
            'state': [self.state],
            'date': [self.date],
            'year': [self.year],
            'office': [self.office],
            'notes': [self.notes],
            'unique_id': [self.unique_id]
        })

        self._parsed_cvr = self._prepare_parsed_cvr(parser_func=parser_func,
                                                    parser_args=parser_args,
                                                    parsed_cvr=parsed_cvr)
        self._modified_cvrs = {}
        self._candidate_sets = {}
        self._rule_sets = {}

        # make a default rule set that is just the parsed cvr
        self._default_rule_set_name = '__cvr'
        self.add_rule_set(self._default_rule_set_name, BallotMarks.new_rule_set())

        # STAT INFO

        self._cvr_stat_table = None
        self._compute_cvr_stat_table()

        self._summary_cvr_stat_table = None
        self._compute_summary_cvr_stat_table()

        self._split_filter_dict = {}
        self._summary_cvr_split_stat_table = None

    # CVR MODS
    def _prepare_parsed_cvr(self,
                            parser_func: Optional[Callable] = None,
                            parser_args: Optional[Dict] = None,
                            parsed_cvr: Optional[Dict[str, List]] = None) -> Dict[str, List]:

        if parser_func and parser_args:
            parsed_cvr = parser_func(**parser_args)

        if not parsed_cvr:
            raise ValueError('if no parser_func and parser_args are passed, a parsed_cvr must be passed.')

        # if parser returns a list, assume it is rank list of lists
        if isinstance(parsed_cvr, list):
            parsed_cvr = {
                'ranks': parsed_cvr,
                'weight': [decimal.Decimal('1')] * len(parsed_cvr)
            }

        if 'ranks' not in parsed_cvr:
            raise RuntimeError('Parsed CVR does not contain field "ranks"')

        if len(parsed_cvr['ranks']) == 0:
            raise RuntimeError('parsed ranks list is empty.')

        if 'weight' not in parsed_cvr:
            parsed_cvr['weight'] = [decimal.Decimal('1') for _ in parsed_cvr['ranks']]

        if not isinstance(parsed_cvr['weight'][0], decimal.Decimal):
            parsed_cvr['weight'] = [decimal.Decimal(str(i)) for i in parsed_cvr['weight']]

        parsed_cvr['ballot_marks'] = [BallotMarks(ranks) for ranks in parsed_cvr['ranks']]
        del parsed_cvr['ranks']

        ballot_lengths = collections.Counter(len(b.marks) for b in parsed_cvr['ballot_marks'])
        if len(ballot_lengths) > 1:
            raise RuntimeError(f'Parsed CVR contains ballots with unequal length rank lists. {str(ballot_lengths)}')

        field_lengths = {k: len(parsed_cvr[k]) for k in parsed_cvr}
        if len(set(field_lengths.values())) > 1:
            raise RuntimeError(f'Parsed CVR contains fields of unequal length. {str(field_lengths)}')

        return parsed_cvr

    def _unique_id(self) -> str:
        pieces = []
        if self.jurisdiction:
            pieces.append(self.jurisdiction)
        if self.date:
            padded_date = "".join(date_piece if len(date_piece) > 1 else "0" + date_piece
                                  for date_piece in self.date.split("/"))
            pieces.append(padded_date)
        elif self.year:
            pieces.append(self.year)
        if self.office:
            pieces.append(self.office)
        return "_".join(re.sub('[^0-9a-zA-Z_]+', '', piece) for piece in pieces)

    def _make_modified_cvr(self, rule_set_name: str) -> None:

        if rule_set_name not in self._rule_sets:
            raise RuntimeError(f'rule set {rule_set_name} has not yet been added using add_rule_set().')

        cvr = {k: v for k, v in self._parsed_cvr.items()}
        cvr['ballot_marks'] = [b.copy() for b in cvr['ballot_marks']]

        for ballot in cvr['ballot_marks']:
            ballot.apply_rules(**self._rule_sets[rule_set_name])

        self._modified_cvrs.update({rule_set_name: cvr})

    def _make_candidate_set(self, rule_set_name: str) -> None:

        if rule_set_name not in self._rule_sets:
            raise RuntimeError(f'rule set {rule_set_name} has not yet been added using add_rule_set().')

        cvr = self._parsed_cvr

        # unpack rules
        rule_set = self._rule_sets[rule_set_name]
        combine_writeins = rule_set['combine_writein_marks']
        exclude_writeins = rule_set['exclude_writein_marks']

        candidate_ballot_marks = BallotMarks(set.union(*[b.unique_candidates for b in cvr['ballot_marks']]))
        candidate_ballot_marks.apply_rules(combine_writein_marks=combine_writeins, exclude_writein_marks=exclude_writeins)

        self._candidate_sets.update({rule_set_name: candidate_ballot_marks})

    def add_rule_set(self, set_name: str, set_dict: Dict[str, Optional[bool]]) -> None:

        # if rule set already created, do nothing
        if set_name in self._rule_sets and set_dict == self._rule_sets[set_name]:
            return

        # if making a new rule set using the same name, delete old one
        if set_name in self._rule_sets and set_dict != self._rule_sets[set_name]:
            del self._modified_cvrs[set_name]
            del self._candidate_sets[set_name]

        self._rule_sets.update({set_name: set_dict})

    def get_cvr_dict(self, rule_set_name: Optional[str] = None) -> Dict[str, List]:

        if rule_set_name is None:
            rule_set_name = self._default_rule_set_name

        if rule_set_name not in self._modified_cvrs:
            self._make_modified_cvr(rule_set_name)

        return self._modified_cvrs[rule_set_name]

    def get_candidates(self, rule_set_name: Optional[str] = None) -> BallotMarks:

        if rule_set_name is None:
            rule_set_name = self._default_rule_set_name

        if rule_set_name not in self._candidate_sets:
            self._make_candidate_set(rule_set_name)

        return self._candidate_sets[rule_set_name]




import pytest

from rcv_cruncher.cvr.base import CastVoteRecord

# public methods, no custom errors reachable
# add_rule_set
# get_cvr_dict
# get_candidates
# get_cvr_stats
# get_cvr_split_stats

params = [
    (ValueError, None),
    (RuntimeError, {'ranks': []}),
    (RuntimeError, {'RANKS': [['A', 'B', 'C']]}),
    (RuntimeError, {'ranks': [['A'], ['A', 'B']]}),
    (RuntimeError, {'ranks': [['A', 'B'], ['A', 'B']], 'weight': [1, 1, 1]})
]


@pytest.mark.parametrize("error_type, inputs", params)
def test_constructor_errors(error_type, inputs):

    with pytest.raises(error_type):
        CastVoteRecord(parsed_cvr=inputs)


def test_get_cvr_table_errors():

    cvr = CastVoteRecord(parsed_cvr={
        'ranks': [["A", "B", "C"]]
    })

    with pytest.raises(RuntimeError):
        cvr.cvr_table(table_format=None)
